import pytest
import requests.exceptions

from iceprod.roles_groups import ROLES, GROUP_PRIORITIES


async def test_rest_auth_roles(server):
    client = server(roles=['admin'])
    ret = await client.request('GET', '/roles')
    assert ret['results'] == list(ROLES.keys())

    client = server(roles=['user'])
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('GET', '/roles')
    assert exc_info.value.response.status_code == 403


async def test_rest_auth_groups(server):
    client = server(roles=['admin'])
    ret = await client.request('GET', '/groups')
    assert len(ret['results']) == len(GROUP_PRIORITIES)
    assert [x['name'] for x in ret['results']] == list(GROUP_PRIORITIES.keys())
    assert [x['priority'] for x in ret['results']] == list(GROUP_PRIORITIES.values())

    client = server(roles=['user'])
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('GET', '/groups')
    assert exc_info.value.response.status_code == 403


async def test_rest_auth_users(server):
    client = server(roles=['admin'])
    ret = await client.request('GET', '/users')
    assert ret['results'] == []

    await client.request('PUT', '/users/foo')

    ret = await client.request('GET', '/users')
    assert len(ret['results']) == 1
    assert ret['results'][0]['username'] == 'foo'
    assert 'priority' in ret['results'][0]

    ret = await client.request('GET', '/users/foo')
    assert ret['username'] == 'foo'
    assert 'priority' in ret

    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('PUT', '/users/foo')
    assert exc_info.value.response.status_code == 409

    await client.request('DELETE', '/users/foo')
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('GET', '/users/foo')
    assert exc_info.value.response.status_code == 404


async def test_rest_auth_users_bad_role(server):
    client = server(roles=['user'])
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('GET', '/users')
    assert exc_info.value.response.status_code == 403

    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('GET', '/users/foo')
    assert exc_info.value.response.status_code == 403

    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('PUT', '/users/foo')
    assert exc_info.value.response.status_code == 403

    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('DELETE', '/users/foo')
    assert exc_info.value.response.status_code == 403

    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('PUT', '/users/foo/priority')
    assert exc_info.value.response.status_code == 403


async def test_rest_auth_users_from_dataset(server):
    client = server(username='foo', roles=['user'], groups=['users', 'simprod'])

    # create dataset, which should set user prio
    data = {
        'description': 'blah',
        'tasks_per_job': 4,
        'jobs_submitted': 1,
        'tasks_submitted': 4,
        'group': 'users',
    }
    await client.request('POST', '/datasets', data)

    client = server(roles=['admin'])
    ret = await client.request('GET', '/users')
    assert len(ret['results']) == 1
    assert ret['results'][0]['username'] == 'foo'
    assert 'priority' in ret['results'][0]


async def test_rest_auth_user_priority(server):
    client = server(roles=['admin'])
    await client.request('PUT', '/users/foo')
    await client.request('PUT', '/users/foo/priority', {'priority': 0.3})

    ret = await client.request('GET', '/users/foo')
    assert ret['username'] == 'foo'
    assert ret['priority'] == 0.3

    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('PUT', '/users/foo/priority', {'priority': -1})
    assert exc_info.value.response.status_code == 400

    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('PUT', '/users/foo/priority', {'priority': 10})
    assert exc_info.value.response.status_code == 400

    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('PUT', '/users/foo/priority', {'priority': 'str'})
    assert exc_info.value.response.status_code == 400

    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('PUT', '/users/foo/priority')
    assert exc_info.value.response.status_code == 400

    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('PUT', '/users/bar/priority', {'priority': 0.5})
    assert exc_info.value.response.status_code == 400


async def test_rest_auth_auths(server):
    client = server(username='me', roles=['user'], groups=['users'])

    # create dataset first
    data = {
        'description': 'blah',
        'tasks_per_job': 4,
        'jobs_submitted': 1,
        'tasks_submitted': 4,
        'group': 'users',
    }
    ret = await client.request('POST', '/datasets', data)
    dataset_id = ret['result']

    # user can't do this
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', '/auths', {})
    assert exc_info.value.response.status_code == 403

    client = server(roles=['system'])
    args = {
        'name': 'dataset_id',
        'value': dataset_id,
        'role': 'read',
        'username': 'me',
        'groups': [],
    }
    await client.request('POST', '/auths', args)

    args = {
        'name': 'dataset_id',
        'value': dataset_id,
        'role': 'read',
        'username': 'you',
        'groups': ['users'],
    }
    await client.request('POST', '/auths', args)

    args = {
        'name': 'dataset_id',
        'value': dataset_id,
        'role': 'read',
        'username': 'you',
        'groups': ['foo'],
    }
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', '/auths', args)
    assert exc_info.value.response.status_code == 403

    # now check for write auth
    args = {
        'name': 'dataset_id',
        'value': dataset_id,
        'role': 'write',
        'username': 'me',
        'groups': [],
    }
    await client.request('POST', '/auths', args)

    args = {
        'name': 'dataset_id',
        'value': dataset_id,
        'role': 'write',
        'username': 'you',
        'groups': ['users'],
    }
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', '/auths', args)
    assert exc_info.value.response.status_code == 403

    args = {
        'name': 'dataset_id',
        'value': dataset_id,
        'role': 'write',
        'username': 'you',
        'groups': ['foo'],
    }
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', '/auths', args)
    assert exc_info.value.response.status_code == 403

    # bad dataset_id
    args = {
        'name': 'dataset_id',
        'value': 'foo',
        'role': 'read',
        'username': 'me',
        'groups': [],
    }
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', '/auths', args)
    assert exc_info.value.response.status_code == 403

    # bad role
    args = {
        'name': 'dataset_id',
        'value': dataset_id,
        'role': 'foo',
        'username': 'me',
        'groups': [],
    }
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', '/auths', args)
    assert exc_info.value.response.status_code == 403



"""
Test script for dataclasses
"""

from __future__ import absolute_import, division, print_function

from tests.util import unittest_reporter, glob_tests

import logging
logger = logging.getLogger('dataclasses')


try:
    pass
except:
    pass

import unittest
import iceprod.core.dataclasses


class dataclasses_test(unittest.TestCase):
    def setUp(self):
        super(dataclasses_test,self).setUp()

    def tearDown(self):
        super(dataclasses_test,self).tearDown()

    @unittest_reporter
    def test_01_Job(self):
        """Test the Job class"""
        j = iceprod.core.dataclasses.Job()

        if not j.valid():
            raise Exception('empty job not valid')

        j.convert()

        if not j.valid():
            raise Exception('converted empty job not valid')

    @unittest_reporter
    def test_02_Steering(self):
        """Test the Steering class"""
        s = iceprod.core.dataclasses.Steering()

        if not s.valid():
            raise Exception('empty steering not valid')

        s.convert()

        if not s.valid():
            raise Exception('converted empty steering not valid')

    @unittest_reporter
    def test_03_Task(self):
        """Test the Task class"""
        t = iceprod.core.dataclasses.Task()

        if not t.valid():
            raise Exception('empty task not valid')

        t.convert()

        if not t.valid():
            raise Exception('converted empty task not valid')

    @unittest_reporter
    def test_04_Tray(self):
        """Test the Tray class"""
        t = iceprod.core.dataclasses.Tray()

        if not t.valid():
            raise Exception('empty tray not valid')

        t.convert()

        if not t.valid():
            raise Exception('converted empty tray not valid')

    @unittest_reporter
    def test_05_Module(self):
        """Test the Module class"""
        m = iceprod.core.dataclasses.Module()

        if not m.valid():
            raise Exception('empty module not valid')

        m.convert()

        if not m.valid():
            raise Exception('converted empty module not valid')

    @unittest_reporter
    def test_06_Class(self):
        """Test the Class class"""
        c = iceprod.core.dataclasses.Class()

        if not c.valid():
            raise Exception('empty class not valid')

        c.convert()

        if not c.valid():
            raise Exception('converted empty class not valid')

    @unittest_reporter
    def test_08_Resource(self):
        """Test the Resource class"""
        r = iceprod.core.dataclasses.Resource()

        if not r.valid():
            raise Exception('empty resource not valid')

        r.convert()

        if not r.valid():
            raise Exception('converted empty resource not valid')

        r['transfer'] = False
        self.assertIs(r.do_transfer(), False)
        r['transfer'] = 'f'
        self.assertIs(r.do_transfer(), False)
        r['transfer'] = 'N'
        self.assertIs(r.do_transfer(), False)
        r['transfer'] = 0
        self.assertIs(r.do_transfer(), False)
        r['transfer'] = True
        self.assertIs(r.do_transfer(), True)
        r['transfer'] = 'T'
        self.assertIs(r.do_transfer(), True)
        r['transfer'] = 'Yes'
        self.assertIs(r.do_transfer(), True)
        r['transfer'] = 1
        self.assertIs(r.do_transfer(), True)
        r['transfer'] = 'maybe'
        self.assertEqual(r.do_transfer(), 'maybe')
        r['transfer'] = 'If'
        self.assertEqual(r.do_transfer(), 'maybe')
        r['transfer'] = 'if needed'
        self.assertEqual(r.do_transfer(), 'maybe')
        r['transfer'] = 'exists'
        self.assertEqual(r.do_transfer(), 'maybe')
        r['transfer'] = 'blah'
        self.assertIs(r.do_transfer(), True)
        r['transfer'] = 1234
        self.assertIs(r.do_transfer(), True)
        r['transfer'] = [1,2,3]
        self.assertIs(r.do_transfer(), True)

    @unittest_reporter
    def test_09_Data(self):
        """Test the Data class"""
        d = iceprod.core.dataclasses.Data()

        if not d.valid():
            raise Exception('empty data not valid')

        d.convert()

        if not d.valid():
            raise Exception('converted empty data not valid')

    @unittest_reporter
    def test_10_Batchsys(self):
        """Test the Batchsys class"""
        b = iceprod.core.dataclasses.Batchsys()

        if not b.valid():
            raise Exception('empty batchsys not valid')

        b.convert()

        if not b.valid():
            raise Exception('converted empty batchsys not valid')

    @unittest_reporter
    def test_20_DifPlus(self):
        """Test the DifPlus class"""
        d = iceprod.core.dataclasses.DifPlus()

        if not d.valid():
            raise Exception('empty difplus not valid')

        d.convert()

        if not d.valid():
            raise Exception('converted empty difplus not valid')

    @unittest_reporter
    def test_21_Dif(self):
        """Test the Dif class"""
        d = iceprod.core.dataclasses.Dif()

        if not d.valid():
            raise Exception('empty dif not valid')

        d.convert()

        if not d.valid():
            raise Exception('converted empty dif not valid')

    @unittest_reporter
    def test_22_Plus(self):
        """Test the Plus class"""
        p = iceprod.core.dataclasses.Plus()

        if not p.valid():
            raise Exception('empty plus not valid')

        p.convert()

        if not p.valid():
            raise Exception('converted empty plus not valid')

    @unittest_reporter
    def test_23_Personnel(self):
        """Test the Personnel class"""
        p = iceprod.core.dataclasses.Personnel()

        if not p.valid():
            raise Exception('empty personnel not valid')

        p.convert()

        if not p.valid():
            raise Exception('converted empty personnel not valid')

    @unittest_reporter
    def test_24_Datacenter(self):
        """Test the DataCenter class"""
        d = iceprod.core.dataclasses.DataCenter()

        if not d.valid():
            raise Exception('empty datacenter not valid')

        d.convert()

        if not d.valid():
            raise Exception('converted empty datacenter not valid')


def load_tests(loader, tests, pattern):
    suite = unittest.TestSuite()
    alltests = glob_tests(loader.getTestCaseNames(dataclasses_test))
    suite.addTests(loader.loadTestsFromNames(alltests,dataclasses_test))
    return suite



import logging

from cachetools.func import ttl_cache
import tornado.web

from iceprod.roles_groups import GROUPS
from .base import authenticated, PublicHandler

logger = logging.getLogger('website-dataset')


class DatasetBrowse(PublicHandler):
    """Handle /dataset urls"""
    @ttl_cache(maxsize=256, ttl=600)
    async def get_usernames(self):
        ret = await self.system_rest_client.request('GET', '/users')
        return [x['username'] for x in ret['results']]

    @authenticated
    async def get(self):
        assert self.rest_client
        usernames = await self.get_usernames()
        filter_options = {
            'status': ['processing', 'suspended', 'errors', 'complete', 'truncated'],
            'groups': list(GROUPS.keys()),
            'users': usernames,
        }
        filter_results = {n: self.get_arguments(n) for n in filter_options}

        args = {'keys': 'dataset_id|dataset|status|description|group|username'}
        for name in filter_results:
            val = filter_results[name]
            if not val:
                continue
            if any(v not in filter_options[name] for v in val):
                raise tornado.web.HTTPError(400, reason='Bad filter '+name+' value')
            args[name] = '|'.join(val)

        url = '/datasets'

        ret = await self.rest_client.request('GET', url, args)
        datasets = sorted(ret.values(), key=lambda x:x.get('dataset',0), reverse=True)
        logger.debug('datasets: %r', datasets)
        datasets = filter(lambda x: 'dataset' in x, datasets)
        self.render(
            'dataset_browse.html',
            datasets=datasets,
            filter_options=filter_options,
            filter_results=filter_results,
        )


class Dataset(PublicHandler):
    """Handle /dataset urls"""
    @authenticated
    async def get(self, dataset_id):
        assert self.rest_client
        if dataset_id.isdigit():
            try:
                d_num = int(dataset_id)
                if d_num < 10000000:
                    all_datasets = await self.rest_client.request('GET', '/datasets', {'keys': 'dataset_id|dataset'})
                    for d in all_datasets.values():
                        if d['dataset'] == d_num:
                            dataset_id = d['dataset_id']
                            break
            except Exception:
                pass
        try:
            dataset = await self.rest_client.request('GET','/datasets/{}'.format(dataset_id))
        except Exception:
            raise tornado.web.HTTPError(404, reason='Dataset not found')
        dataset_num = dataset['dataset']

        passkey = self.auth_access_token

        jobs = await self.rest_client.request('GET','/datasets/{}/job_counts/status'.format(dataset_id))
        tasks = await self.rest_client.request('GET','/datasets/{}/task_counts/status'.format(dataset_id))
        task_info = await self.rest_client.request('GET','/datasets/{}/task_counts/name_status'.format(dataset_id))
        task_stats = await self.rest_client.request('GET','/datasets/{}/task_stats'.format(dataset_id))
        try:
            config = await self.rest_client.request('GET','/config/{}'.format(dataset_id))
        except Exception:
            config = {}
        for t in task_info:
            logger.info('task_info[%s] = %r', t, task_info[t])
            type_ = 'UNK'
            for task in config.get('tasks', []):
                if 'name' in task and task['name'] == t:
                    type_ = 'GPU' if 'requirements' in task and 'gpu' in task['requirements'] and task['requirements']['gpu'] else 'CPU'
                    break
            task_info[t] = {
                'name': t,
                'type': type_,
                'idle': task_info[t].get('idle', 0),
                'waiting': task_info[t].get('waiting', 0),
                'queued': task_info[t].get('queued', 0),
                'running': task_info[t].get('processing', 0),
                'complete': task_info[t].get('complete', 0),
                'error': task_info[t].get('failed', 0) + task_info[t].get('suspended', 0),
            }
        self.render('dataset_detail.html', dataset_id=dataset_id, dataset_num=dataset_num,
                    dataset=dataset, jobs=jobs, tasks=tasks, task_info=task_info,
                    task_stats=task_stats, passkey=passkey)



"""
Credentials store and refresh.
"""
import asyncio
import dataclasses
from datetime import datetime, UTC
import logging
import time
from typing import Any, Self

import jwt
from prometheus_client import Info, start_http_server
import pymongo
import pymongo.errors
import requests.exceptions
from rest_tools.client import RestClient, ClientCredentialsAuth
from rest_tools.server import RestServer, ArgumentHandler, ArgumentSource
from tornado.web import HTTPError
from tornado.web import RequestHandler as TornadoRequestHandler
from wipac_dev_tools import from_environment_as_dataclass

from iceprod.util import VERSION_STRING
from iceprod.common.mongo import Mongo
from iceprod.common.prom_utils import AsyncMonitor
from iceprod.rest.auth import authorization
from iceprod.rest.base_handler import IceProdRestConfig, APIBase
from iceprod.server.util import nowstr, datetime2str
from .service import ExchangeException, RefreshService
from .util import get_expiration, is_expired

logger = logging.getLogger('server')


class BaseCredentialsHandler(APIBase):
    def initialize(self, *args, refresh_service: RefreshService, rest_client: RestClient, **kwargs):  # type: ignore[override]
        super().initialize(*args, **kwargs)
        self.refresh_service = refresh_service
        self.rest_client = rest_client

    async def check_attr_auth(self, arg, val, role):
        """
        Based on the request groups or username, check if they are allowed to
        access `arg`:`role`.

        Runs a remote query to the IceProd API.

        Args:
            arg (str): attribute name to check
            val (str): attribute value
            role (str): the role to check for (read|write)
        """
        assert self.rest_client
        args = {
            'name': arg,
            'value': val,
            'role': role,
            'username': self.current_user,
            'groups': self.auth_groups,
        }
        try:
            await self.rest_client.request('POST', '/auths', args)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 403:
                raise HTTPError(403, 'auth failed')
            else:
                raise HTTPError(500, 'auth could not be completed')

    async def create(self, db, base_data):
        assert self.refresh_service
        now = time.time()
        argo = ArgumentHandler(ArgumentSource.JSON_BODY_ARGUMENTS, self)
        argo.add_argument('url', type=str, required=True)
        argo.add_argument('type', type=str, choices=['s3', 'oauth'], required=True)
        argo.add_argument('transfer_prefix', type=str, default='', required=False)
        argo.add_argument('buckets', type=list, default=[], required=False)
        argo.add_argument('access_key', type=str, default='', required=False)
        argo.add_argument('secret_key', type=str, default='', required=False)
        argo.add_argument('access_token', type=str, default='', required=False)
        argo.add_argument('refresh_token', type=str, default='', required=False)
        argo.add_argument('scope', type=str, default=None, required=False)
        argo.add_argument('expiration', type=float, default=now, required=False)
        argo.add_argument('last_use', type=float, default=now, required=False)
        args = vars(argo.parse_args())
        url = args['url']
        credential_type = args['type']

        base_data['url'] = url
        base_data['transfer_prefix'] = args['transfer_prefix']
        data = base_data.copy()
        data.update({
            'type': credential_type,
        })

        if credential_type == 's3':
            if not args['buckets']:
                raise HTTPError(400, reason='must specify bucket(s)')
            if not args['access_key']:
                raise HTTPError(400, reason='must specify access_key')
            if not args['secret_key']:
                raise HTTPError(400, reason='must specify secret_key')

            data['buckets'] = args['buckets']
            data['access_key'] = args['access_key']
            data['secret_key'] = args['secret_key']

        elif credential_type == 'oauth':
            if (not args['access_token']) and not args['refresh_token']:
                raise HTTPError(400, reason='must specify either access or refresh tokens')
            data['access_token'] = args['access_token']
            data['refresh_token'] = args['refresh_token']
            if args['scope']:
                base_data['scope'] = args['scope']
            data['scope'] = args['scope']
            data['expiration'] = args['expiration']
            if data['access_token'] and data.get('expiration') == now:
                data['expiration'] = get_expiration(data['access_token'])
            data['last_use'] = args['last_use']
            if data['access_token'] and data['scope'] is None:
                data['scope'] = jwt.decode(data['access_token'], options={"verify_signature": False}).get('scope', '')

            if 'refresh_token' in data and not data['access_token']:
                new_cred = await self.refresh_service.refresh_cred(data)
                data.update(new_cred)

        else:
            raise HTTPError(400, reason='bad credential type')

        try:
            await db.update_one(
                base_data,
                {'$set': data},
                upsert=True,
            )
        except pymongo.errors.DuplicateKeyError:
            raise HTTPError(409, reason='credential already exists')

    async def patch_cred(self, db, base_data):
        assert self.refresh_service
        argo = ArgumentHandler(ArgumentSource.JSON_BODY_ARGUMENTS, self)
        argo.add_argument('url', type=str, required=True)
        argo.add_argument('type', type=str, choices=['s3', 'oauth'], required=True)
        argo.add_argument('transfer_prefix', type=str, default='', required=False)
        argo.add_argument('buckets', type=list, default=[], required=False)
        argo.add_argument('access_key', type=str, default='', required=False)
        argo.add_argument('secret_key', type=str, default='', required=False)
        argo.add_argument('access_token', type=str, default='', required=False)
        argo.add_argument('refresh_token', type=str, default='', required=False)
        argo.add_argument('scope', type=str, default='', required=False)
        argo.add_argument('expiration', type=float, default=0, required=False)
        argo.add_argument('last_use', type=float, default=0, required=False)
        args = vars(argo.parse_args())
        base_data['url'] = args['url']
        base_data['type'] = args['type']
        if args['transfer_prefix']:
            base_data['transfer_prefix'] = args['transfer_prefix']

        data = {}
        for key in ('buckets', 'access_key', 'secret_key', 'access_token', 'refresh_token', 'scope', 'expiration', 'last_use'):
            if val := args[key]:
                data[key] = val

        if 'refresh_token' in data and 'access_token' not in data:
            new_cred = await self.refresh_service.refresh_cred(data)
            data.update(new_cred)
        if 'access_token' in data and 'expiration' not in data:
            data['expiration'] = get_expiration(data['access_token'])

        ret = await db.find_one_and_update(
            base_data,
            {'$set': data},
        )
        if not ret:
            raise HTTPError(404, 'credential not found')

    async def search_creds(self, db, base_data):
        assert self.refresh_service
        if url := self.get_argument('url', None):
            base_data['url'] = url
        if transfer_prefix := self.get_argument('transfer_prefix', None):
            base_data['transfer_prefix'] = transfer_prefix
        if scope := self.get_argument('scope', None):
            base_data['scope'] = scope
        logger.info('base_data: %r', base_data)

        refresh = self.get_argument('norefresh', None) is None

        if refresh:
            update_data = {'last_use': time.time()}
            filters = base_data.copy()
            filters['type'] = 'oauth'
            await db.update_many(filters, {'$set': update_data})

        ret = []
        async for row in db.find(base_data, projection={'_id': False}):
            if refresh and is_expired(row) and row['refresh_token']:
                try:
                    new_cred = await self.refresh_service.refresh_cred(row)
                    filters = base_data.copy()
                    filters['url'] = row['url']
                    cred = await db.find_one_and_update(filters, {'$set': new_cred}, projection={'_id': False})
                    ret.append(cred)
                except Exception:
                    logging.debug('ignore expired token %r', row)
            else:
                ret.append(row)

        return ret

    async def delete_cred(self, db, base_data):
        assert self.refresh_service
        if url := self.get_argument('url', None):
            base_data['url'] = url
        if transfer_prefix := self.get_argument('transfer_prefix', None):
            base_data['transfer_prefix'] = transfer_prefix
        if scope := self.get_argument('scope', None):
            base_data['scope'] = scope
        logger.info('base_data: %r', base_data)

        await db.delete_many(base_data)

    async def exchange_cred(self, db, base_data):
        assert self.refresh_service

        client_id = self.get_argument('client_id')

        # must be oauth for exchange
        base_data['type'] = 'oauth'

        new_scope = self.get_argument('new_scope', None)

        # get unexpired tokens
        creds = await self.search_creds(db, base_data)

        ret = []
        for row in creds:
            try:
                c = await self.refresh_service.exchange_cred(row, client_id=client_id, new_scope=new_scope)
            except ExchangeException:
                continue
            ret.append(c)
        logger.info('exchange_cred found %d creds and exchanged %d creds', len(creds), len(ret))
        return ret


class CreateHandler(BaseCredentialsHandler):
    """
    Handle requests for a new credential.  Just return it, don't store it.
    """
    @authorization(roles=['admin', 'system', 'user'])
    async def post(self):
        """
        Generate an oauth credential

        Args:
            username (str): the username for the credential
            url (str): url of controlled resource
            transfer_prefix (str): transfer prefix for file transfer
            scope (str): scope of access token
        """
        username = self.get_argument('username')
        scope = self.get_argument('scope')
        url = self.get_argument('url')
        transfer_prefix = self.get_argument('transfer_prefix')
        if self.auth_roles == ['user'] and username != self.current_user:
            raise HTTPError(403, 'unauthorized')

        try:
            new_cred = await self.refresh_service.create_cred(
                url=url,
                transfer_prefix=transfer_prefix,
                username=username,
                scope=scope
            )
        except Exception as e:
            raise HTTPError(400, reason=str(e))
        self.write(new_cred)


class GroupCredentialsHandler(BaseCredentialsHandler):
    """
    Handle group credentials requests.
    """
    @authorization(roles=['admin', 'system', 'user'])
    async def get(self, groupname):
        """
        Get a group's credentials.

        Args:
            groupname (str): group name
        Param args:
            url (str): (optional) url of controlled resource
            transfer_prefix (str): transfer prefix for file transfer
            scope (str): (optional) scope of access token
        Returns:
            dict: url: credential dict
        """
        if self.auth_roles == ['user'] and groupname not in self.auth_groups:
            raise HTTPError(403, 'unauthorized')

        ret = await self.search_creds(self.db.group_creds, {'groupname': groupname})
        self.write(ret)

    @authorization(roles=['admin', 'system', 'user'])
    async def post(self, groupname):
        """
        Set a group credential.  Overwrites an existing credential for the specified url.

        Common body args:
            url (str): url of controlled resource
            transfer_prefix (str): transfer prefix for file transfer
            type (str): credential type (`s3` or `oauth`)

        S3 body args:
            buckets (list): list of buckets for this url, or [] if using virtual-hosted buckets in the url
            access_key (str): access key
            secret_key (str): secret key

        OAuth body args:
            access_token (str): access token
            refresh_token (str): refresh token
            scope (str): scope of access token
            expiration (float): access token expiration, in unix time (optional)

        Args:
            groupname (str): group name
        """
        if self.auth_roles == ['user'] and groupname not in self.auth_groups:
            raise HTTPError(403, 'unauthorized')

        await self.create(self.db.group_creds, {'groupname': groupname})
        self.write({})

    @authorization(roles=['admin', 'system', 'user'])
    async def patch(self, groupname):
        """
        Update a group credential.  Usually used to update a specifc field.

        Body args:
            url (str): url of controlled resource
            transfer_prefix (str): transfer prefix for file transfer
            scope (str): (optional) scope of access token

        Other body args will update a credential.

        Args:
            groupname (str): group name
        """
        if self.auth_roles == ['user'] and groupname not in self.auth_groups:
            raise HTTPError(403, 'unauthorized')

        await self.patch_cred(self.db.group_creds, {'groupname': groupname})
        self.write({})

    @authorization(roles=['admin', 'system', 'user'])
    async def delete(self, groupname):
        """
        Delete a group's credentials.

        Args:
            groupname (str): groupname
        Param args:
            url (str): (optional) url of controlled resource
            transfer_prefix (str): transfer prefix for file transfer
            scope (str): (optional) scope of access token
        """
        if self.auth_roles == ['user'] and groupname not in self.auth_groups:
            raise HTTPError(403, 'unauthorized')

        await self.delete_cred(self.db.group_creds, {'groupname': groupname})
        self.write({})


class GroupExchangeHandler(BaseCredentialsHandler):
    """
    Handle group exchange requests.
    """
    @authorization(roles=['admin', 'system', 'user'])
    async def get(self, groupname):
        """
        Get a group's exchange credentials.

        Args:
            groupname (str): group name
        Param args:
            url (str): (optional) url of controlled resource
            transfer_prefix (str): transfer prefix for file transfer
            scope (str): (optional) scope of access token
            client_id (str): client_id to exchange to
        Returns:
            dict: url: credential dict
        """
        if self.auth_roles == ['user'] and groupname not in self.auth_groups:
            raise HTTPError(403, 'unauthorized')

        ret = await self.exchange_cred(self.db.group_creds, {'groupname': groupname})
        self.write(ret)


class UserCredentialsHandler(BaseCredentialsHandler):
    """
    Handle user credentials requests.
    """
    @authorization(roles=['admin', 'system', 'user'])
    async def get(self, username):
        """
        Get a user's credentials.

        Args:
            username (str): username
        Param args:
            url (str): (optional) url of controlled resource
            transfer_prefix (str): transfer prefix for file transfer
            scope (str): (optional) scope of access token
        Returns:
            dict: url: credential dict
        """
        if self.auth_roles == ['user'] and username != self.current_user:
            raise HTTPError(403, 'unauthorized')

        ret = await self.search_creds(self.db.user_creds, {'username': username})
        self.write(ret)

    @authorization(roles=['admin', 'system', 'user'])
    async def post(self, username):
        """
        Set a user credential.  Overwrites an existing credential for the specified url.

        Common body args:
            url (str): url of controlled resource
            transfer_prefix (str): transfer prefix for file transfer
            type (str): credential type (`s3` or `oauth`)

        S3 body args:
            buckets (list): list of buckets for this url, or [] if using virtual-hosted buckets in the url
            access_key (str): access key
            secret_key (str): secret key

        OAuth body args:
            access_token (str): access token
            refresh_token (str): refresh token
            scope (str): scope of access token
            expiration (float): access token expiration, in unix time (optional)

        Args:
            username (str): username
        """
        if self.auth_roles == ['user'] and username != self.current_user:
            raise HTTPError(403, 'unauthorized')

        await self.create(self.db.user_creds, {'username': username})
        self.write({})

    @authorization(roles=['admin', 'system', 'user'])
    async def patch(self, username):
        """
        Update a user credential.  Usually used to update a specifc field.

        Body args:
            url (str): url of controlled resource
            transfer_prefix (str): transfer prefix for file transfer
            scope (str): (optional) scope of access token

        Other body args will update a credential.

        Args:
            username (str): username
        """
        if self.auth_roles == ['user'] and username != self.current_user:
            raise HTTPError(403, 'unauthorized')

        await self.patch_cred(self.db.user_creds, {'username': username})
        self.write({})

    @authorization(roles=['admin', 'system', 'user'])
    async def delete(self, username):
        """
        Delete a user's credentials.

        Args:
            username (str): username
        Param args:
            url (str): (optional) url of controlled resource
            transfer_prefix (str): transfer prefix for file transfer
            scope (str): (optional) scope of access token
        Returns:
            dict: url: credential dict
        """
        if self.auth_roles == ['user'] and username != self.current_user:
            raise HTTPError(403, 'unauthorized')

        await self.delete_cred(self.db.user_creds, {'username': username})
        self.write({})


class UserExchangeHandler(BaseCredentialsHandler):
    """
    Handle user exchange requests.
    """
    @authorization(roles=['admin', 'system', 'user'])
    async def get(self, username):
        """
        Get a user's exchange credentials.

        Args:
            username (str): user name
        Param args:
            url (str): (optional) url of controlled resource
            transfer_prefix (str): transfer prefix for file transfer
            scope (str): (optional) scope of access token
            client_id (str): client_id to exchange to
        Returns:
            dict: url: credential dict
        """
        if self.auth_roles == ['user'] and username != self.current_user:
            raise HTTPError(403, 'unauthorized')

        ret = await self.exchange_cred(self.db.user_creds, {'username': username})
        self.write(ret)


class DatasetCredentialsHandler(BaseCredentialsHandler):
    """
    Handle dataset credentials requests.
    """
    @authorization(roles=['admin', 'system'])
    async def get(self, dataset_id):
        """
        Get a datasets's credentials.

        Args:
            dataset_id (str): dataset_id
        Param args:
            url (str): (optional) url of controlled resource
            transfer_prefix (str): transfer prefix for file transfer
            scope (str): (optional) scope of access token
        Returns:
            dict: url: credential dict
        """
        ret = await self.search_creds(self.db.dataset_creds, {'dataset_id': dataset_id})
        self.write(ret)

    @authorization(roles=['admin', 'system'])
    async def post(self, dataset_id):
        """
        Set a dataset credential.  Overwrites an existing credential for the specified url.

        Common body args:
            url (str): url of controlled resource
            transfer_prefix (str): transfer prefix for file transfer
            type (str): credential type (`s3` or `oauth`)

        S3 body args:
            buckets (list): list of buckets for this url, or [] if using virtual-hosted buckets in the url
            access_key (str): access key
            secret_key (str): secret key

        OAuth body args:
            access_token (str): access token
            refresh_token (str): refresh token
            scope (str): scope of access token
            expiration (float): access token expiration, in unix time (optional)

        Args:
            dataset_id (str): dataset_id
        """
        await self.create(self.db.dataset_creds, {'dataset_id': dataset_id})
        self.write({})

    @authorization(roles=['admin', 'system'])
    async def patch(self, dataset_id):
        """
        Update a dataset credential.  Usually used to update a specifc field.

        Body args:
            url (str): url of controlled resource
            transfer_prefix (str): transfer prefix for file transfer
            scope (str): (optional) scope of access token

        Other body args will update a credential.

        Args:
            dataset_id (str): dataset_id
        """
        await self.patch_cred(self.db.dataset_creds, {'dataset_id': dataset_id})
        self.write({})

    @authorization(roles=['admin', 'system'])
    async def delete(self, dataset_id):
        """
        Delete a dataset's credentials.

        Args:
            dataset_id (str): dataset_id
        Param args:
            url (str): (optional) url of controlled resource
            transfer_prefix (str): transfer prefix for file transfer
            scope (str): (optional) scope of access token
        Returns:
            dict: url: credential dict
        """
        await self.delete_cred(self.db.dataset_creds, {'dataset_id': dataset_id})
        self.write({})


class DatasetExchangeHandler(BaseCredentialsHandler):
    """
    Handle dataset exchange requests.
    """
    @authorization(roles=['admin', 'system'])
    async def get(self, dataset_id):
        """
        Get a dataset's exchange credentials.

        Args:
            dataset_id (str): dataset_id
        Param args:
            url (str): (optional) url of controlled resource
            transfer_prefix (str): transfer prefix for file transfer
            scope (str): (optional) scope of access token
            client_id (str): client_id to exchange to
        Returns:
            dict: url: credential dict
        """
        ret = await self.exchange_cred(self.db.dataset_creds, {'dataset_id': dataset_id})
        self.write(ret)


class DatasetTaskCredentialsHandler(BaseCredentialsHandler):
    """
    Handle dataset/task credentials requests.
    """
    @authorization(roles=['admin', 'system'])
    async def get(self, dataset_id, task_name):
        """
        Get a datasets's credentials.

        Args:
            dataset_id (str): dataset_id
            task_name (str): task name
        Param args:
            url (str): (optional) url of controlled resource
            transfer_prefix (str): transfer prefix for file transfer
            scope (str): (optional) scope of access token
        Returns:
            dict: url: credential dict
        """
        ret = await self.search_creds(self.db.dataset_creds, {'dataset_id': dataset_id, 'task_name': task_name})
        self.write(ret)

    @authorization(roles=['admin', 'system'])
    async def post(self, dataset_id, task_name):
        """
        Set a dataset credential.  Overwrites an existing credential for the specified url.

        Common body args:
            url (str): url of controlled resource
            transfer_prefix (str): transfer prefix for file transfer
            type (str): credential type (`s3` or `oauth`)

        S3 body args:
            buckets (list): list of buckets for this url, or [] if using virtual-hosted buckets in the url
            access_key (str): access key
            secret_key (str): secret key

        OAuth body args:
            access_token (str): access token
            refresh_token (str): refresh token
            scope (str): scope of access token
            expiration (float): access token expiration, in unix time (optional)

        Args:
            dataset_id (str): dataset_id
            task_name (str): task name
        """
        await self.create(self.db.dataset_creds, {'dataset_id': dataset_id, 'task_name': task_name})
        self.write({})

    @authorization(roles=['admin', 'system'])
    async def patch(self, dataset_id, task_name):
        """
        Update a dataset credential.  Usually used to update a specifc field.

        Body args:
            url (str): url of controlled resource
            transfer_prefix (str): transfer prefix for file transfer
            scope (str): (optional) scope of access token

        Other body args will update a credential.

        Args:
            dataset_id (str): dataset_id
            task_name (str): task name
        """
        await self.patch_cred(self.db.dataset_creds, {'dataset_id': dataset_id, 'task_name': task_name})
        self.write({})

    @authorization(roles=['admin', 'system'])
    async def delete(self, dataset_id, task_name):
        """
        Delete a dataset's credentials.

        Args:
            dataset_id (str): dataset_id
            task_name (str): task name
        Param args:
            url (str): (optional) url of controlled resource
            transfer_prefix (str): transfer prefix for file transfer
            scope (str): (optional) scope of access token
        Returns:
            dict: url: credential dict
        """
        await self.delete_cred(self.db.dataset_creds, {'dataset_id': dataset_id, 'task_name': task_name})
        self.write({})


class DatasetTaskExchangeHandler(BaseCredentialsHandler):
    """
    Handle dataset task exchange requests.
    """
    @authorization(roles=['admin', 'system'])
    async def get(self, dataset_id, task_name):
        """
        Get a dataset's exchange credentials.

        Args:
            dataset_id (str): dataset_id
            task_name (str): task name
        Param args:
            url (str): (optional) url of controlled resource
            transfer_prefix (str): transfer prefix for file transfer
            scope (str): (optional) scope of access token
            client_id (str): client_id to exchange to
        Returns:
            dict: url: credential dict
        """
        ret = await self.exchange_cred(self.db.dataset_creds, {'dataset_id': dataset_id, 'task_name': task_name})
        self.write(ret)


class HealthHandler(BaseCredentialsHandler):
    """
    Handle health requests.
    """
    async def get(self):
        """
        Get health status.

        Returns based on exit code, 200 = ok, 400 = failure
        """
        assert self.refresh_service
        now = time.time()
        status = {
            'now': nowstr(),
            'start_time': datetime2str(datetime.fromtimestamp(self.refresh_service.start_time, tz=UTC)),
            'last_run_time': "",
            'last_success_time': "",
        }
        try:
            if self.refresh_service.last_run_time is None and self.refresh_service.start_time + 3600 < now:
                self.send_error(500, reason='refresh was never run')
                return
            if self.refresh_service.last_run_time is not None:
                if self.refresh_service.last_run_time + 3600 < now:
                    self.send_error(500, reason='refresh has stopped running')
                    return
                status['last_run_time'] = datetime2str(datetime.utcfromtimestamp(self.refresh_service.last_run_time))
            if self.refresh_service.last_success_time is None and self.refresh_service.start_time + 86400 < now:
                self.send_error(500, reason='refresh was never successful')
                return
            if self.refresh_service.last_success_time is not None:
                if self.refresh_service.last_success_time + 86400 < now:
                    self.send_error(500, reason='refresh has stopped being successful')
                    return
                status['last_success_time'] = datetime2str(datetime.utcfromtimestamp(self.refresh_service.last_success_time))
        except Exception:
            logger.info('error from refresh service', exc_info=True)
            self.send_error(500, reason='error from refresh service')
            return

        self.write(status)


class Error(TornadoRequestHandler):
    def prepare(self):
        raise HTTPError(404, 'invalid api route')


@dataclasses.dataclass
class DefaultConfig:
    HOST: str = 'localhost'
    PORT: int = 8080
    DEBUG: bool = False
    OPENID_URL: str = ''
    OPENID_AUDIENCE: str = ''
    ICEPROD_API_ADDRESS: str = 'https://api.iceprod.icecube.aq'
    ICEPROD_API_CLIENT_ID: str = ''
    ICEPROD_API_CLIENT_SECRET: str = ''
    TOKEN_CLIENTS: str = '{}'
    TOKEN_REFRESH_WINDOW: float = 168.0  # hours
    TOKEN_EXPIRE_BUFFER: int = 15  # minutes
    TOKEN_SERVICE_CHECK_INTERVAL: int = 180
    DB_URL: str = 'mongodb://localhost/datasets'
    DB_TIMEOUT: int = 60
    DB_WRITE_CONCERN: int = 1
    PROMETHEUS_PORT: int = 0
    CI_TESTING: str = ''


class Server:
    def __init__(self: Self):
        config = from_environment_as_dataclass(DefaultConfig)

        rest_config: dict[str, Any] = {
            'debug': config.DEBUG,
        }
        if config.OPENID_URL:
            logging.info(f'enabling auth via {config.OPENID_URL} for aud "{config.OPENID_AUDIENCE}"')
            rest_config.update({
                'auth': {
                    'openid_url': config.OPENID_URL,
                    'audience': config.OPENID_AUDIENCE,
                }
            })
        elif config.CI_TESTING:
            rest_config.update({
                'auth': {
                    'secret': 'secret',
                }
            })
        else:
            raise RuntimeError('OPENID_URL not specified, and CI_TESTING not enabled!')

        # enable monitoring
        self.prometheus_port = config.PROMETHEUS_PORT if config.PROMETHEUS_PORT > 0 else None
        self.async_monitor = None

        self.db_client = Mongo(
            url=config.DB_URL,
            timeout=config.DB_TIMEOUT,
            write_concern=config.DB_WRITE_CONCERN
        )
        self.indexes = {
            'group_creds': {
                'group_index2': {'keys': [('groupname', pymongo.DESCENDING), ('transfer_prefix', pymongo.DESCENDING)], 'unique': True},
            },
            'user_creds': {
                'username_index2': {'keys': [('username', pymongo.DESCENDING), ('transfer_prefix', pymongo.DESCENDING)], 'unique': True},
            },
            'dataset_creds': {
                'dataset_index2': {
                    'keys': [
                        ('dataset_id', pymongo.DESCENDING),
                        ('task_name', pymongo.DESCENDING),
                        ('transfer_prefix', pymongo.DESCENDING)
                    ],
                    'unique': True,
                },
            }
        }

        rest_client: RestClient
        if config.ICEPROD_API_CLIENT_ID and config.ICEPROD_API_CLIENT_SECRET:
            logging.info(f'enabling auth via {config.OPENID_URL} for aud "{config.OPENID_AUDIENCE}"')
            rest_client = ClientCredentialsAuth(
                address=config.ICEPROD_API_ADDRESS,
                token_url=config.OPENID_URL,
                client_id=config.ICEPROD_API_CLIENT_ID,
                client_secret=config.ICEPROD_API_CLIENT_SECRET,
            )
        elif config.CI_TESTING:
            rest_client = RestClient(config.ICEPROD_API_ADDRESS, timeout=1, retries=0)
        else:
            raise RuntimeError('ICEPROD_API_CLIENT_ID or ICEPROD_API_CLIENT_SECRET not specified, and CI_TESTING not enabled!')

        self.refresh_service = RefreshService(
            database=self.db_client.db,
            clients=config.TOKEN_CLIENTS,
            refresh_window=config.TOKEN_REFRESH_WINDOW,
            expire_buffer=config.TOKEN_EXPIRE_BUFFER,
            service_run_interval=config.TOKEN_SERVICE_CHECK_INTERVAL,
        )
        self.refresh_service_task = None

        kwargs = IceProdRestConfig(rest_config, database=self.db_client.db)
        kwargs['refresh_service'] = self.refresh_service
        kwargs['rest_client'] = rest_client

        server = RestServer(debug=config.DEBUG)

        server.add_route('/create', CreateHandler, kwargs)
        server.add_route(r'/groups/(?P<groupname>\w+)/credentials', GroupCredentialsHandler, kwargs)
        server.add_route(r'/groups/(?P<groupname>\w+)/exchange', GroupExchangeHandler, kwargs)
        server.add_route(r'/users/(?P<username>\w+)/credentials', UserCredentialsHandler, kwargs)
        server.add_route(r'/users/(?P<username>\w+)/exchange', UserExchangeHandler, kwargs)
        server.add_route(r'/datasets/(?P<dataset_id>\w+)/credentials', DatasetCredentialsHandler, kwargs)
        server.add_route(r'/datasets/(?P<dataset_id>\w+)/exchange', DatasetExchangeHandler, kwargs)
        server.add_route(r'/datasets/(?P<dataset_id>\w+)/tasks/(?P<task_name>[^\/\?\#]+)/credentials', DatasetTaskCredentialsHandler, kwargs)
        server.add_route(r'/datasets/(?P<dataset_id>\w+)/tasks/(?P<task_name>[^\/\?\#]+)/exchange', DatasetTaskExchangeHandler, kwargs)
        server.add_route('/healthz', HealthHandler, kwargs)
        server.add_route(r'/(.*)', Error)

        server.startup(address=config.HOST, port=config.PORT)

        self.server = server

    async def start(self):
        if self.prometheus_port:
            logging.info("starting prometheus on %r", self.prometheus_port)
            start_http_server(self.prometheus_port)
            i = Info('iceprod', 'IceProd information')
            i.info({
                'version': VERSION_STRING,
                'type': 'credentials',
            })
            self.async_monitor = AsyncMonitor(labels={'type': 'credentials'})
            await self.async_monitor.start()

        await self.db_client.ping()
        await self.db_client.create_indexes(indexes=self.indexes)

        if not self.refresh_service_task:
            self.refresh_service_task = asyncio.create_task(self.refresh_service.run())

    async def stop(self):
        await self.server.stop()
        if self.refresh_service_task:
            self.refresh_service_task.cancel()
            try:
                await self.refresh_service_task
            except asyncio.CancelledError:
                pass  # ignore cancellations
            finally:
                self.refresh_service_task = None
        if self.async_monitor:
            await self.async_monitor.stop()
        await self.db_client.close()



"""
Test script for scheduled_tasks/update_task_priority
"""

import logging
from unittest.mock import AsyncMock, MagicMock

import pytest
from iceprod.scheduled_tasks import update_task_priority

logger = logging.getLogger('scheduled_tasks_update_task_priority_test')


async def test_200_run():
    rc = MagicMock()
    async def client(method, url, args=None):
        logger.info('REST: %s, %s', method, url)
        if url == '/tasks/bar':
            client.called = True
            return {}
        elif url == '/tasks':
            return {'tasks':[{'task_id':'bar','dataset_id':'foo'}]}
        elif url == '/datasets':
            return {'foo':{'dataset_id':'foo','start_date':'2024-01-01T01:00:00','username':'a','group':'g','tasks_submitted':200,'jobs_submitted':100,'priority':1.}}
        elif url == '/datasets/foo/tasks':
            return {'bar':{'task_id':'bar','dataset_id':'foo','task_index':0,'job_index':12}}
        elif url == '/datasets/foo/tasks/bar':
            return {'task_id':'bar','dataset_id':'foo','task_index':0,'job_index':12}
        elif url == '/users':
            return {'results':[{'username':'a','priority':0.5}]}
        elif url == '/groups':
            return {'results':[{'name':'g','priority':0.5}]}
        else:
            raise Exception()
    client.called = False
    rc.request = client

    await update_task_priority.run(rc, status=['idle', 'waiting'], debug=True)
    assert client.called


async def test_201_run():
    rc = MagicMock()
    pilots = {'a':{}}
    # try tasks error
    rc.request = AsyncMock(side_effect=Exception())

    # check it normally hides the error
    await update_task_priority.run(rc, status=['idle', 'waiting'], debug=False)

    with pytest.raises(Exception):
        await reset_tasks.run(rc, debug=True)




"""
Task Queue
==========

Utilities for fairly queueing tasks based on resource usage.
"""

from __future__ import absolute_import, division, print_function

import logging


logger = logging.getLogger('task_queue')


def sched_prio(resources, time_in_queue=0):
    """
    Weight the priority to prefer smaller resource requests.
    As a task stays longer in the queue, increase priority.

    Best priority is 0, worse increases to infinity.

    Args:
        resources (dict): resources dict
        time_in_queue (float): the time this task has spent in the queue

    Returns:
        float: priority
    """
    return max(0, sum(resources.values()) - time_in_queue/600)


def get_queue(resources):
    """
    Determine which queue this task belongs in.

    Args:
        resources (dict): resources dict

    Returns:
        str: queue name
    """
    if resources['gpu']:
        return 'gpu'
    elif resources['memory'] >= 4:
        return 'memory'
    else:
        return 'default'



import logging
import time
from unittest.mock import AsyncMock

import pytest
import pytest_asyncio
from rest_tools.client import RestClient
from rest_tools.utils import Auth
import requests.exceptions

from iceprod.credentials.service import ExchangeException
from iceprod.server.util import nowstr
from iceprod.rest.auth import ROLES, GROUPS
from iceprod.credentials.server import Server


@pytest_asyncio.fixture
async def server(monkeypatch, port, mongo_url, mongo_clear):
    monkeypatch.setenv('CI_TESTING', '1')
    monkeypatch.setenv('PORT', str(port))
    monkeypatch.setenv('ICEPROD_API_ADDRESS', 'http://test.iceprod')

    s = Server()
    # don't call start, because we don't want the service to run
    # await s.start()

    s.refresh_service.refresh_cred = AsyncMock(return_value={})
    s.refresh_service.exchange_cred = AsyncMock(return_value={})
    s.refresh_service.create_cred = AsyncMock(return_value={})

    auth = Auth('secret')

    def _add_to_data(data, attrs):
        logging.debug('attrs: %r', attrs)
        for item in attrs:
            key,value = item.split('=',1)
            d = data
            while '.' in key:
                k,key = key.split('.',1)
                if k not in d:
                    d[k] = {}
                d = d[k]
            if key not in d:
                d[key] = []
            d[key].append(value)

    def client(timeout=1, username='user', roles=[], groups=[], exp=10):
        data = {'preferred_username': username}
        for r in roles:
            _add_to_data(data, ROLES[r])
        for g in groups:
            _add_to_data(data, GROUPS[g])
        token = auth.create_token('username', expiration=exp, payload=data)
        return RestClient(f'http://localhost:{port}', token, timeout=timeout, retries=0)

    try:
        yield client, s.refresh_service
    finally:
        await s.stop()


GROUP = 'simprod'
USER = 'username'
DATASETS = ['dataset1', 'dataset2']
TASK_NAMES = ['alpha', 'bravo']


async def test_credentials_create(server):
    client = server[0](roles=['system'])
    refresh = server[1]

    refresh.create_cred.return_value = {
        'foo': 'bar'
    }

    data = {
        'username': 'foo',
        'scope': 'storage.modify:/foo/bar storage.read:/',
        'url': 'http://foo',
        'transfer_prefix': 'foo://',
    }
    ret = await client.request('POST', f'/create', data)
    assert ret == refresh.create_cred.return_value
    assert refresh.create_cred.call_args.kwargs == data


async def test_credentials_groups_empty(server):
    client = server[0](roles=['system'])

    ret = await client.request('GET', f'/groups/{GROUP}/credentials')
    assert ret == []


async def test_credentials_groups_s3(server):
    client = server[0](roles=['system'])

    data = {
        'url': 'http://foo',
        'type': 's3',
        'access_key': 'XXXX',
        'secret_key': 'YYYY',
        'buckets': ['bar'],
    }
    await client.request('POST', f'/groups/{GROUP}/credentials', data)

    ret = await client.request('GET', f'/groups/{GROUP}/credentials')
    data['groupname'] = GROUP
    data['transfer_prefix'] = ''
    assert ret == [data]

    # test bucket in url
    data2 = {
        'url': 'http://bucket.bar',
        'type': 's3',
        'access_key': 'XXXX',
        'secret_key': 'YYYY',
        'buckets': ['bucket'],
    }
    await client.request('POST', f'/groups/{GROUP}/credentials', data2)

    ret = await client.request('GET', f'/groups/{GROUP}/credentials')
    data2['groupname'] = GROUP
    data2['transfer_prefix'] = ''
    assert ret == [data, data2]

    # now overwrite
    data3 = {
        'url': 'http://foo',
        'type': 's3',
        'access_key': 'XXXX',
        'secret_key': 'YYYY',
        'buckets': ['baz'],
    }
    await client.request('POST', f'/groups/{GROUP}/credentials', data3)

    ret = await client.request('GET', f'/groups/{GROUP}/credentials')
    data3_out = data3.copy()
    data3_out['groupname'] = GROUP
    data3_out['transfer_prefix'] = ''
    assert ret == [data3_out, data2]

    await client.request('DELETE', f'/groups/{GROUP}/credentials', {'url': 'http://foo'})

    ret = await client.request('GET', f'/groups/{GROUP}/credentials')
    assert ret == [data2]

    await client.request('POST', f'/groups/{GROUP}/credentials', data3)
    await client.request('DELETE', f'/groups/{GROUP}/credentials')

    ret = await client.request('GET', f'/groups/{GROUP}/credentials')
    assert ret == []
    
async def test_credentials_groups_s3_bad(server):
    client = server[0](roles=['system'])

    data = {
        'url': 'http://foo',
        'type': 's3',
        'access_key': 'XXXX',
        'secret_key': 'YYYY',
        'buckets': [],
    }
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', f'/groups/{GROUP}/credentials', data)
    assert exc_info.value.response.status_code == 400

    data = {
        'url': 'http://foo',
        'type': 's3',
        'access_key': 'XXXX',
        'secret_key': 'YYYY',
    }
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', f'/groups/{GROUP}/credentials', data)
    assert exc_info.value.response.status_code == 400

    data = {
        'url': 'http://foo',
        'type': 's3',
        'secret_key': 'YYYY',
        'buckets': ['bar'],
    }
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', f'/groups/{GROUP}/credentials', data)
    assert exc_info.value.response.status_code == 400

    data = {
        'url': 'http://foo',
        'type': 's3',
        'access_key': 'XXXX',
        'buckets': ['bar'],
    }
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', f'/groups/{GROUP}/credentials', data)
    assert exc_info.value.response.status_code == 400


async def test_credentials_groups_oauth(server):
    client = server[0](roles=['system'])

    data = {
        'url': 'http://foo',
        'type': 'oauth',
        'access_token': str(client.access_token),
    }
    await client.request('POST', f'/groups/{GROUP}/credentials', data)

    ret = await client.request('GET', f'/groups/{GROUP}/credentials')
    for k in data:
        assert ret[0][k] == data[k]


async def test_credentials_groups_user(server):
    client = server[0](roles=['user'], groups=['users', GROUP])

    data = {
        'url': 'http://foo',
        'type': 'oauth',
        'access_token': str(client.access_token),
    }
    await client.request('POST', f'/groups/{GROUP}/credentials', data)

    ret = await client.request('GET', f'/groups/{GROUP}/credentials')
    for k in data:
        assert ret[0][k] == data[k]

    data2 = {
        'url': 'http://foo',
        'type': 'oauth',
        'refresh_token': str(client.access_token),
        'expiration': time.time()+100,
    }
    await client.request('PATCH', f'/groups/{GROUP}/credentials', data2)
    ret = await client.request('GET', f'/groups/{GROUP}/credentials')
    assert ret[0]['refresh_token'] == data2['refresh_token']

    await client.request('DELETE', f'/groups/{GROUP}/credentials')

    # bad user
    client = server[0](roles=['user'], groups=['users'])

    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', f'/groups/{GROUP}/credentials', data)
    assert exc_info.value.response.status_code == 403

    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('GET', f'/groups/{GROUP}/credentials')
    assert exc_info.value.response.status_code == 403

    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('DELETE', f'/groups/{GROUP}/credentials')
    assert exc_info.value.response.status_code == 403


async def test_credentials_users_empty(server):
    client = server[0](roles=['system'])

    ret = await client.request('GET', f'/users/{USER}/credentials')
    assert ret == []


async def test_credentials_users_s3(server):
    client = server[0](roles=['system'])

    data = {
        'url': 'http://foo',
        'type': 's3',
        'access_key': 'XXXX',
        'secret_key': 'YYYY',
        'buckets': ['bar'],
    }
    await client.request('POST', f'/users/{USER}/credentials', data)

    ret = await client.request('GET', f'/users/{USER}/credentials')
    data['username'] = USER
    data['transfer_prefix'] = ''
    assert ret == [data]

    data2 = {
        'url': 'http://bar',
        'type': 's3',
        'access_key': 'XXXX',
        'secret_key': 'YYYY',
        'buckets': ['bar'],
    }
    await client.request('POST', f'/users/{USER}/credentials', data2)

    ret = await client.request('GET', f'/users/{USER}/credentials')
    data2['username'] = USER
    data2['transfer_prefix'] = ''
    assert ret == [data, data2]

    # now overwrite
    data3 = {
        'url': 'http://foo',
        'type': 's3',
        'access_key': 'XXXX',
        'secret_key': 'YYYY',
        'buckets': ['baz'],
    }
    await client.request('POST', f'/users/{USER}/credentials', data3)

    ret = await client.request('GET', f'/users/{USER}/credentials')
    data3_out = data3.copy()
    data3_out['username'] = USER
    data3_out['transfer_prefix'] = ''
    assert ret == [data3_out, data2]

    await client.request('DELETE', f'/users/{USER}/credentials', {'url': 'http://foo'})

    ret = await client.request('GET', f'/users/{USER}/credentials')
    assert ret == [data2]

    await client.request('POST', f'/users/{USER}/credentials', data3)
    await client.request('DELETE', f'/users/{USER}/credentials')

    ret = await client.request('GET', f'/users/{USER}/credentials')
    assert ret == []


async def test_credentials_users_oauth(server):
    client = server[0](roles=['system'])

    data = {
        'url': 'http://foo',
        'type': 'oauth',
        'access_token': str(client.access_token),
    }
    await client.request('POST', f'/users/{USER}/credentials', data)

    ret = await client.request('GET', f'/users/{USER}/credentials')
    for k in data:
        assert ret[0][k] == data[k]


async def test_credentials_users_user(server):
    client = server[0](username=USER, roles=['user'], groups=['users'])

    data = {
        'url': 'http://foo',
        'type': 'oauth',
        'access_token': str(client.access_token),
    }
    await client.request('POST', f'/users/{USER}/credentials', data)

    ret = await client.request('GET', f'/users/{USER}/credentials')
    for k in data:
        assert ret[0][k] == data[k]

    data2 = {
        'url': 'http://foo',
        'type': 'oauth',
        'access_token': str(client.access_token),
        'last_use': time.time()
    }
    await client.request('PATCH', f'/users/{USER}/credentials', data2)
    ret = await client.request('GET', f'/users/{USER}/credentials', {'norefresh': 'true'})
    assert ret[0]['last_use'] == data2['last_use']

    await client.request('DELETE', f'/users/{USER}/credentials')

    # bad user
    client = server[0](username='foo', roles=['user'], groups=['users'])
    
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', f'/users/{USER}/credentials', data)
    assert exc_info.value.response.status_code == 403

    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('GET', f'/users/{USER}/credentials')
    assert exc_info.value.response.status_code == 403

    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('DELETE', f'/users/{USER}/credentials')
    assert exc_info.value.response.status_code == 403


async def test_credentials_datasets_user_fail(server):
    client = server[0](roles=['user'], groups=['users', GROUP])
    
    data = {
        'url': 'http://foo',
        'type': 'oauth',
    }
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('GET', f'/datasets/{DATASETS[0]}/credentials', data)
    assert exc_info.value.response.status_code == 403

    data = {
        'url': 'http://foo',
        'type': 'oauth',
        'access_token': str(client.access_token),
    }
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', f'/datasets/{DATASETS[0]}/tasks/{TASK_NAMES[0]}/credentials', data)
    assert exc_info.value.response.status_code == 403


async def test_credentials_datasets(server):
    client = server[0](roles=['system'])
    auth = Auth('secret')

    data = {
        'url': 'http://foo',
        'type': 'oauth',
        'access_token': str(client.access_token),
    }
    await client.request('POST', f'/datasets/{DATASETS[0]}/tasks/{TASK_NAMES[0]}/credentials', data)

    ret = await client.request('GET', f'/datasets/{DATASETS[0]}/tasks/{TASK_NAMES[0]}/credentials')
    for k in data:
        assert ret[0][k] == data[k]

    ret = await client.request('GET', f'/datasets/{DATASETS[0]}/credentials')
    for k in data:
        assert ret[0][k] == data[k]

    data2 = {
        'url': 'http://foo',
        'type': 'oauth',
        'access_token': str(client.access_token),
        'last_use': time.time()
    }
    await client.request('PATCH', f'/datasets/{DATASETS[0]}/tasks/{TASK_NAMES[0]}/credentials', data2)
    ret = await client.request('GET', f'/datasets/{DATASETS[0]}/tasks/{TASK_NAMES[0]}/credentials', {'norefresh': 'true'})
    assert ret[0]['last_use'] == data2['last_use']

    # now for the second task
    data3 = {
        'url': 'http://foo',
        'type': 'oauth',
        'access_token': auth.create_token('sub2'),
    }
    await client.request('POST', f'/datasets/{DATASETS[0]}/tasks/{TASK_NAMES[1]}/credentials', data3)

    ret = await client.request('GET', f'/datasets/{DATASETS[0]}/tasks/{TASK_NAMES[0]}/credentials')
    assert len(ret) == 1
    assert ret[0]['access_token'] == data2['access_token']

    ret = await client.request('GET', f'/datasets/{DATASETS[0]}/tasks/{TASK_NAMES[1]}/credentials')
    assert len(ret) == 1
    assert ret[0]['access_token'] == data3['access_token']

    ret = await client.request('GET', f'/datasets/{DATASETS[0]}/credentials')
    assert len(ret) == 2
    assert ret[0]['access_token'] == data2['access_token']
    assert ret[1]['access_token'] == data3['access_token']

    # now for a different dataset
    data4 = {
        'url': 'http://foo',
        'type': 'oauth',
        'access_token': auth.create_token('sub3'),
    }
    await client.request('POST', f'/datasets/{DATASETS[1]}/tasks/{TASK_NAMES[1]}/credentials', data4)

    ret = await client.request('GET', f'/datasets/{DATASETS[1]}/tasks/{TASK_NAMES[0]}/credentials')
    assert len(ret) == 0

    ret = await client.request('GET', f'/datasets/{DATASETS[1]}/tasks/{TASK_NAMES[1]}/credentials')
    assert len(ret) == 1
    assert ret[0]['access_token'] == data4['access_token']

    ret = await client.request('GET', f'/datasets/{DATASETS[1]}/credentials')
    assert len(ret) == 1
    assert ret[0]['access_token'] == data4['access_token']
    
    # delete
    await client.request('DELETE', f'/datasets/{DATASETS[0]}/credentials')
    ret = await client.request('GET', f'/datasets/{DATASETS[0]}/credentials')
    assert len(ret) == 0
    
    ret = await client.request('GET', f'/datasets/{DATASETS[1]}/credentials')
    assert len(ret) == 1
    assert ret[0]['access_token'] == data4['access_token']

    await client.request('DELETE', f'/datasets/{DATASETS[1]}/credentials')
    ret = await client.request('GET', f'/datasets/{DATASETS[0]}/credentials')
    assert len(ret) == 0
    ret = await client.request('GET', f'/datasets/{DATASETS[1]}/credentials')
    assert len(ret) == 0


async def test_credentials_datasets_empty(server):
    client = server[0](roles=['system'])

    ret = await client.request('GET', f'/datasets/{DATASETS[0]}/credentials')
    assert ret == []


async def test_credentials_datasets_s3(server):
    client = server[0](roles=['system'])

    data = {
        'url': 'http://foo',
        'transfer_prefix': 'http://bar',
        'type': 's3',
        'access_key': 'XXXX',
        'secret_key': 'YYYY',
        'buckets': ['bar'],
    }
    await client.request('POST', f'/datasets/{DATASETS[0]}/tasks/{TASK_NAMES[0]}/credentials', data)

    ret = await client.request('GET', f'/datasets/{DATASETS[0]}/credentials')
    data['dataset_id'] = DATASETS[0]
    data['task_name'] = TASK_NAMES[0]
    assert ret == [data]

    data2 = {
        'url': 'http://bar',
        'transfer_prefix': 'http://bar',
        'type': 's3',
        'access_key': 'XXXX',
        'secret_key': 'YYYY',
        'buckets': ['bar'],
    }
    await client.request('POST', f'/datasets/{DATASETS[0]}/tasks/{TASK_NAMES[0]}/credentials', data2)

    ret = await client.request('GET', f'/datasets/{DATASETS[0]}/tasks/{TASK_NAMES[0]}/credentials')
    data2['dataset_id'] = DATASETS[0]
    data2['task_name'] = TASK_NAMES[0]
    assert ret == [data, data2]

    data4 = {
        'url': 'http://foo',
        'transfer_prefix': 'http://baz',
        'type': 's3',
        'access_key': 'XXXX',
        'secret_key': 'YYYY',
        'buckets': ['bar'],
    }
    await client.request('POST', f'/datasets/{DATASETS[0]}/tasks/{TASK_NAMES[0]}/credentials', data4)

    ret = await client.request('GET', f'/datasets/{DATASETS[0]}/tasks/{TASK_NAMES[0]}/credentials')
    data4['dataset_id'] = DATASETS[0]
    data4['task_name'] = TASK_NAMES[0]
    assert ret == [data, data2, data4]

    # now overwrite
    data3 = {
        'url': 'http://foo',
        'transfer_prefix': 'http://bar',
        'type': 's3',
        'access_key': 'XXXX',
        'secret_key': 'YYYY',
        'buckets': ['baz'],
    }
    await client.request('POST', f'/datasets/{DATASETS[0]}/tasks/{TASK_NAMES[0]}/credentials', data3)

    ret = await client.request('GET', f'/datasets/{DATASETS[0]}/tasks/{TASK_NAMES[0]}/credentials')
    data3_out = data3.copy()
    data3_out['dataset_id'] = DATASETS[0]
    data3_out['task_name'] = TASK_NAMES[0]
    assert ret == [data3_out, data2, data4]

    await client.request('DELETE', f'/datasets/{DATASETS[0]}/tasks/{TASK_NAMES[0]}/credentials', {'url': 'http://foo'})

    ret = await client.request('GET', f'/datasets/{DATASETS[0]}/tasks/{TASK_NAMES[0]}/credentials')
    assert ret == [data2]

    await client.request('POST', f'/datasets/{DATASETS[0]}/tasks/{TASK_NAMES[0]}/credentials', data3)
    await client.request('DELETE', f'/datasets/{DATASETS[0]}/tasks/{TASK_NAMES[0]}/credentials')

    ret = await client.request('GET', f'/datasets/{DATASETS[0]}/tasks/{TASK_NAMES[0]}/credentials')
    assert ret == []


async def test_credentials_datasets_oauth(server):
    client = server[0](roles=['system'])

    data = {
        'url': 'http://foo',
        'type': 'oauth',
        'access_token': str(client.access_token),
    }
    await client.request('POST', f'/datasets/{DATASETS[0]}/tasks/{TASK_NAMES[0]}/credentials', data)

    ret = await client.request('GET', f'/datasets/{DATASETS[0]}/tasks/{TASK_NAMES[0]}/credentials')
    for k in data:
        assert ret[0][k] == data[k]
        

async def test_credentials_dataset_task_exchange(server):
    client = server[0](roles=['system'])
    refresh = server[1]

    refresh.exchange_cred.return_value = {
        'new_cred': 'true'
    }

    data = {
        'url': 'http://foo',
        'type': 'oauth',
        'access_token': str(client.access_token),
        'refresh_token': 'xxxxxxxxx'
    }
    await client.request('POST', f'/datasets/{DATASETS[0]}/tasks/{TASK_NAMES[0]}/credentials', data)

    data['client_id'] = 'my_client'
    ret = await client.request('GET', f'/datasets/{DATASETS[0]}/tasks/{TASK_NAMES[0]}/exchange', data)

    assert ret == [refresh.exchange_cred.return_value]
        

async def test_credentials_dataset_exchange_downscope(server):
    client = server[0](roles=['system'])
    refresh = server[1]

    auth = Auth('secret')
    token1 = auth.create_token('tok1', payload={'scope': 'foo bar baz'})
    token1a = auth.create_token('tok1', payload={'scope': 'foo'})
    token2 = auth.create_token('tok1', payload={'scope': 'baz'})

    def do_exchange(cred, *args, **kwargs):
        if cred['access_token'] == token1:
            return {'access_token': token1a}
        else:
            raise ExchangeException('cannot exchange')
    refresh.exchange_cred.side_effect = do_exchange

    data = {
        'url': 'http://foo',
        'type': 'oauth',
        'access_token': token1,
        'refresh_token': 'xxxxxxxxx',
        'scope': 'foo bar baz',
    }
    await client.request('POST', f'/datasets/{DATASETS[0]}/credentials', data)

    data = {
        'url': 'http://foo',
        'type': 'oauth',
        'access_token': token2,
        'refresh_token': 'xxxxxxxxx',
        'scope': 'baz',
    }
    await client.request('POST', f'/datasets/{DATASETS[0]}/credentials', data)

    args = {
        'client_id': 'my_client',
        'new_scope': 'foo'
    }
    ret = await client.request('GET', f'/datasets/{DATASETS[0]}/exchange', args)

    assert ret == [{'access_token': token1a}]


async def test_credentials_healthz(server):
    client = server[0](username=USER, roles=['user'], groups=['users'])

    ret = await client.request('GET', '/healthz')

    now = nowstr()

    assert ret['now'][:16] == now[:16]
    assert ret['start_time'][:16] == now[:16]
    assert ret['last_run_time'] == ''
    assert ret['last_success_time'] == ''

    
async def test_credentials_bad_route(server):
    client = server[0](username=USER, roles=['user'], groups=['users'])

    with pytest.raises(requests.exceptions.HTTPError):
        await client.request('GET', '/foo')
        




import asyncio
from argparse import ArgumentParser
from collections import defaultdict
from functools import partial
import csv
import json
import logging

from iceprod.client_auth import add_auth_to_argparse, create_rest_client


logger = logging.getLogger()


def task_type(name, ord):
    name = name.lower()
    if any(x in name for x in ('generat', 'corsika', 'background', 'topsimulator', 'nugen', 'genie', 'muongun', 'poly')):
        return 'generate'
    elif any(x in name for x in ('prop', 'clsim', 'ppc', 'hits', 'photo', 'server', 'p0')):
        return 'propagate'
    elif any(x in name for x in ('det', 'ic86', 'trigger')):
        return 'detector'
    elif any(x in name for x in ('baseproc', 'offline', 'filter', 'pffilt', 'level2', 'base_proc', 'l1', 'rehyd', 'classifier')):
        return 'offline'
    else:
        logging.info('unknown task: %s, %d', name, ord)
        return 'other'

def determine_group(simprod, iceprod):
    if not simprod:
        simprod = 'Other'
    if simprod == 'ESTES':
        simprod = 'Neutrino Sources'
    elif simprod == 'SnowStorm':
        simprod = 'Low Energy Astro'
    elif simprod == 'General':
        simprod = 'GENERAL'
    if simprod == 'GENERAL' or simprod == 'Other':
        # can we do better?
        if iceprod['group'] == 'filtering':
            return 'i3filter'
        desc = iceprod['description'].lower()
        if 'icetop' in desc or desc.startswith('it '):
            return 'Cosmic Rays'
        elif 'diffuse' in desc or 'cascade' in desc or 'corsika' in desc or 'muongun' in desc or 'dataset 20904' in desc:
            return 'Diffuse'
        elif 'gen2' in desc:
            return 'Gen2'
        elif 'upgrade' in desc or 'pingu' in desc:
            return 'Upgrade'
        elif 'genie' in desc or 'deepcore' in desc or 'low energy' in desc or 'leptoninjector' in desc:
            return 'Low Energy Astro'
        elif 'nugen' in desc or 'neutrino-generator' in desc or 'nutau' in desc:
            return 'Neutrino Sources'

        user = iceprod['username']
        if user == 'mlarson':
            return 'Upgrade'

        logger.info('failed to do better for %s:%s', user, desc)

    return simprod

async def run(rpc, ceph_rpc, simprod_data, outfile):
    datasets = {}
    iceprod_dataset_data = await rpc.request('GET', f'/datasets', {'keys': 'dataset_id|dataset|group|username|description|start_date'})

    existing_paths = set()
    for data in simprod_data:
        dataset_id = data['did']
        if dataset_id == 'd9a980b2e82611ed9ff800505684797b':
            continue
        number = data['number']
        logging.info('simprod - processing %d %s', number, dataset_id)
        logging.debug('iceprod data: %r', iceprod_dataset_data[dataset_id])
        ret = await rpc.request('GET', f'/datasets/{dataset_id}/task_stats')
        tasks = {}
        for i,name in enumerate(ret):
            task = ret[name]
            type_ = task_type(name, i)
            if type_ in tasks:
                tasks[type_]['gpu'] |= task['gpu'] > 0
                tasks[type_]['goodput'] += task['total_hrs']
                tasks[type_]['badput'] += task['total_err_hrs']
            else:
                tasks[type_] = {
                    'gpu': task['gpu'] > 0,
                    'goodput': task['total_hrs'],
                    'badput': task['total_err_hrs'],
                    'storage': 0,
                }

        if data['paths']:
            gen_disk = 0
            filt_disk = 0
            other_disk = 0
            for path in data['paths'].split(';'):
                if path in existing_paths:
                    continue
                existing_paths.add(path)
                try:
                    ret = await ceph_rpc.request('GET', path)
                except Exception:
                    continue
                if 'generated' in path:
                    gen_disk += ret['size']
                elif 'filtered' in path:
                    filt_disk += ret['size']
                else:
                    other_disk += ret['size']
            
            if 'other' in tasks:
                tasks['other']['storage'] += other_disk
                if filt_disk:
                    if 'offline' in tasks:
                        tasks['offline']['storage'] += filt_disk
                    else:
                        tasks['other']['storage'] += filt_disk
                if gen_disk:
                    if 'detector' in tasks:
                        tasks['detector']['storage'] += gen_disk
                    elif 'propagate' in tasks:
                        tasks['propagate']['storage'] += gen_disk
                    elif 'generate' in tasks:
                        tasks['generate']['storage'] += gen_disk
                    elif 'offline' in tasks:
                        tasks['offline']['storage'] += gen_disk
                    else:
                        tasks['other']['storage'] += filt_disk

            else:
                filt_disk += other_disk
                if filt_disk:
                    if 'offline' in tasks:
                        tasks['offline']['storage'] += filt_disk
                    else:
                        gen_disk += filt_disk
                if gen_disk:
                    if 'detector' in tasks:
                        tasks['detector']['storage'] += gen_disk
                    elif 'propagate' in tasks:
                        tasks['propagate']['storage'] += gen_disk
                    elif 'generate' in tasks:
                        tasks['generate']['storage'] += gen_disk
                    elif 'offline' in tasks:
                        tasks['offline']['storage'] += gen_disk
                    else:
                        logging.info('paths %s', data['paths'])
                        logging.info('tasks: %r', list(tasks.keys()))
                        raise Exception('cannot allocate storage to any task!')

        datasets[number] = {
            'dataset_id': dataset_id,
            'tasks': tasks,
            'group': determine_group(simprod=data['group'], iceprod=iceprod_dataset_data[dataset_id]),
            'year': iceprod_dataset_data[dataset_id]['start_date'].split('-',1)[0],
        }
    
    # now go through things not in simprod
    for dataset in iceprod_dataset_data.values():
        dataset_id = dataset['dataset_id']
        number = dataset.get('dataset', -1)
        logging.info('iceprod - processing %d %s', number, dataset_id)
        if number == -1:
            raise Exception()
        if number in datasets:
            continue
        ret = await rpc.request('GET', f'/datasets/{dataset_id}/task_stats')
        tasks = {}
        for i,name in enumerate(ret):
            task = ret[name]
            type_ = task_type(name, i)
            if type_ in tasks:
                tasks[type_]['gpu'] |= task['gpu'] > 0
                tasks[type_]['goodput'] += task['total_hrs']
                tasks[type_]['badput'] += task['total_err_hrs']
            else:
                tasks[type_] = {
                    'gpu': task['gpu'] > 0,
                    'goodput': task['total_hrs'],
                    'badput': task['total_err_hrs'],
                    'storage': 0,
                }
        datasets[number] = {
            'dataset_id': dataset_id,
            'tasks': tasks,
            'group': determine_group(simprod='', iceprod=dataset),
            'year': iceprod_dataset_data[dataset_id]['start_date'].split('-',1)[0],
        }

    group_data = defaultdict(partial(defaultdict, dict))
    for dataset in datasets.values():
        for name in dataset['tasks']:
            task = dataset['tasks'][name]
            is_gpu = task['gpu']
            d = group_data[dataset['group']][dataset['year']]
            if name not in d:
                d[name] = {
                    'cpu': 0.,
                    'gpu': 0.,
                    'storage': 0.,
                }
            d[name]['gpu' if is_gpu else 'cpu'] += task['goodput'] + task['badput']
            d[name]['storage'] += task['storage'] / 1000000000000.  # convert to TB

    with open(outfile, 'w', newline='') as csvfile:
        fieldnames = ['group', 'year', 'task', 'gpu', 'cpu', 'storage']
        writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
        writer.writeheader()
        for group in group_data:
            for year in group_data[group]:
                for task in group_data[group][year]:
                    row = {
                        'group': group,
                        'year': year,
                        'task': task,
                        'gpu': int(group_data[group][year][task]['gpu']),
                        'cpu': int(group_data[group][year][task]['cpu']),
                        'storage': group_data[group][year][task]['storage'],
                    }
                    writer.writerow(row)


def main():
    parser = ArgumentParser()
    parser.add_argument('infile')
    parser.add_argument('outfile')
    parser.add_argument('--log-level', default='INFO')
    add_auth_to_argparse(parser)
    args = parser.parse_args()

    logging.basicConfig(level=getattr(logging, args.log_level.upper()))
    rpc = create_rest_client(args)
    args.rest_url = 'https://disk-usage.icecube.aq/api'
    ceph_rpc = create_rest_client(args, retries=0)

    with open(args.infile, 'r') as f:
        simprod_data = json.load(f)

    asyncio.run(run(rpc, ceph_rpc, simprod_data, args.outfile))


if __name__ == '__main__':
    main()



#!/usr/bin/env python
import os
import sys
import argparse
import logging
import asyncio
import importlib

from rest_tools.client import RestClient

iceprod_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(0,iceprod_path)

logger = logging.getLogger()

def main():
    parser = argparse.ArgumentParser(description='run a scheduled task once')
    parser.add_argument('-t', '--token', help='auth token')
    parser.add_argument('--debug', default=False, type=bool, help='debug enabled/disabled')
    parser.add_argument('--dataset_id', default=None, help='dataset_id')
    parser.add_argument('scheduled_task', help='name of the scheduled task to run')

    args = parser.parse_args()
    args = vars(args)

    logformat='%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'
    logging.basicConfig(format=logformat, level=logging.DEBUG if args['debug'] else logging.INFO)

    rpc = RestClient('https://iceprod2-api.icecube.wisc.edu', args['token'])

    mod = importlib.import_module('iceprod.server.scheduled_tasks.'+args['scheduled_task'])
    fn = getattr(mod,'run')
    kwargs = {'debug': args['debug']}
    if args['dataset_id']:
        kwargs['dataset_id'] = args['dataset_id']

    asyncio.run(fn(rpc, **kwargs))

if __name__ == '__main__':
    main()



import os
import pymongo
from iceprod.common.mongo import Mongo


async def test_mongo_init(mongo_url, mongo_clear):
    url = os.environ['DB_URL']
    m = Mongo(url=url)
    await m.ping()

    db_name = url.rsplit('/',1)[1]
    assert m.client[db_name] == m.db
    assert m.client['bar'] == m['bar']
    await m.close()

async def test_mongo_indexes(mongo_url, mongo_clear):
    url = os.environ['DB_URL']
    m = Mongo(url=url)

    indexes = {
        'bar': {
            'uuid_index': {
                'keys': [
                    ('uuid', pymongo.DESCENDING),
                ],
                'unique': True,
            }
        }
    }
    await m.create_indexes(indexes=indexes)



"""
Test script for globus utilities
"""

from __future__ import absolute_import, division, print_function

from tests.util import unittest_reporter, glob_tests

import logging
logger = logging.getLogger('globus_test')

import os
import sys
import shutil
import subprocess
import tempfile

import unittest

from iceprod.core import to_log
from iceprod.server.globus import SiteGlobusProxy

skip_tests = False
if subprocess.call(['which','grid-proxy-init']):
    skip_tests = True

class siteglobusproxy_test(unittest.TestCase):
    def setUp(self):
        self._timeout = 1
        self.test_dir = tempfile.mkdtemp(dir=os.getcwd())
        if not os.path.exists(self.test_dir):
            os.mkdir(self.test_dir)
        if not skip_tests:
            # clear any proxies
            subprocess.call(['grid-proxy-destroy'],stdout=subprocess.DEVNULL,stderr=subprocess.DEVNULL)
        super(siteglobusproxy_test,self).setUp()

    def tearDown(self):
        shutil.rmtree(self.test_dir)
        super(siteglobusproxy_test,self).tearDown()

    @unittest_reporter(skip=skip_tests, module='globus.SiteGlobusProxy')
    def test_01_init(self):
        cfgfile = os.path.join(self.test_dir,'cfg')
        p = SiteGlobusProxy(cfgfile=cfgfile)
        if not os.path.exists(cfgfile):
            raise Exception('cfgfile does not exist')

    @unittest_reporter(skip=skip_tests, module='globus.SiteGlobusProxy')
    def test_02_init_duration(self):
        cfgfile = os.path.join(self.test_dir,'cfg')
        p = SiteGlobusProxy(cfgfile=cfgfile, duration=10)
        if not os.path.exists(cfgfile):
            raise Exception('cfgfile does not exist')

    @unittest_reporter(skip=skip_tests, module='globus.SiteGlobusProxy')
    def test_10_update_proxy(self):
        cfgfile = os.path.join(self.test_dir,'cfg')
        p = SiteGlobusProxy(cfgfile=cfgfile)
        with self.assertRaises(Exception):
            p.update_proxy()

        p.set_passphrase('gibberish')
        with self.assertRaises(Exception):
            with to_log(sys.stderr), to_log(sys.stdout):
                p.update_proxy()

    @unittest_reporter(name='update_proxy() voms', skip=skip_tests, module='globus.SiteGlobusProxy')
    def test_10_5_update_proxy(self):
        cfgfile = os.path.join(self.test_dir,'cfg')
        p = SiteGlobusProxy(cfgfile=cfgfile)
        p.set_voms_vo('test_vo')
        p.set_voms_role('test_role')
        with self.assertRaises(Exception):
            p.update_proxy()
        p.set_passphrase('gibberish')
        with self.assertRaises(Exception):
            with to_log(sys.stderr), to_log(sys.stdout):
                p.update_proxy()

    @unittest_reporter(skip=skip_tests, module='globus.SiteGlobusProxy')
    def test_11_get_proxy(self):
        cfgfile = os.path.join(self.test_dir,'cfg')
        p = SiteGlobusProxy(cfgfile=cfgfile)
        with self.assertRaises(Exception):
            p.get_proxy()

def load_tests(loader, tests, pattern):
    suite = unittest.TestSuite()
    alltests = glob_tests(loader.getTestCaseNames(siteglobusproxy_test))
    suite.addTests(loader.loadTestsFromNames(alltests,siteglobusproxy_test))
    return suite



import json
import logging
from unittest.mock import AsyncMock, MagicMock, PropertyMock
from urllib.parse import urlparse, parse_qs

from bs4 import BeautifulSoup
import pytest
import requests
import requests_mock

from iceprod.core.config import Config


async def test_website_root(server):
    client = server()
    await client.request('GET', '/')


async def test_website_schemas(server):
    client = server()
    ret = await client.request('GET', '/schemas/v3/dataset.schema.json')
    ret = json.loads(ret)
    assert ret['title'] == 'IceProd Dataset Config'
    
    ret = await client.request('GET', '/schemas/v3/dataset_v3.1.schema.json')
    ret = json.loads(ret)
    assert ret['title'] == 'IceProd Dataset Config'
    logging.info("schema:%r", ret)
    assert ret['properties']['version']['default'] == 3.1
    
    ret = await client.request('GET', '/schemas/v3/dataset_v3.2.schema.json')
    ret = json.loads(ret)
    assert ret['title'] == 'IceProd Dataset Config'
    logging.info("schema:%r", ret)
    assert ret['properties']['version']['default'] == 3.2

    ret = await client.request('GET', '/schemas/v3/config.schema.json')
    ret = json.loads(ret)
    assert ret['title'] == 'IceProd Server Config'


async def test_website_submit(server):
    client = server(username='username', roles=['user'], groups=['users', 'simprod'])

    description = 'Test dataset'

    config = Config({
        'tasks':[{
            'name': 'testing',
            'trays': [{
                'modules': [{
                    'src': '/usr/bin/python3',
                    'args': ''
                }]
            }]
        }],
        'version': 3.2,
    })
    config.fill_defaults()
    config.validate()

    async with client.get_http_client() as http_client:
        ret = await client.request_raw(http_client, 'GET', '/submit')
        ret.raise_for_status()
        #logging.info('ret: %r', ret.text)
        doc = BeautifulSoup(ret.text, 'html.parser')
        xsrf = doc.find('input', {'name': '_xsrf'}).get('value')  # type: ignore
        logging.info('xsrf: %r', xsrf)
        logging.info('cookies: %r', http_client.cookies)

        config_str = json.dumps(config.config, indent=2)

        submit_mock = client.req_mock.add_mock('/actions/submit', {'result': '123'})

        ret = await client.request_raw(http_client, 'POST', '/submit', form_data={
            '_xsrf': xsrf,
            'submit_box': config_str,
            'description': description,
            'number_jobs': 10,
            'group': 'users',
        })
        assert ret.status_code == 303
        assert ret.headers['location'].endswith('/submit/status/123')

        assert submit_mock.called
        assert json.loads(submit_mock.call_args[0][2]['config']) == config.config
        assert submit_mock.call_args[0][2]['description'] == description


def MockError(status_code=400, body=''):
    def _inner(*args, **kwargs):
        response = MagicMock()
        response.status_code = status_code
        if body and isinstance(body, dict):
            response.text = json.dumps(body)
            response.json = MagicMock(return_value=body)
        elif body:
            response.text = body
            response.json = MagicMock(return_value=None)
        raise requests.exceptions.HTTPError(response=response)
    return _inner


async def test_website_submit_invalid(server):
    client = server(username='username', roles=['user'], groups=['users', 'simprod'])

    description = 'Test dataset'

    config = {'tasks':[
        {
            'name': 'testing',
            'trays': []
        }
    ]}

    async with client.get_http_client() as http_client:
        ret = await client.request_raw(http_client, 'GET', '/submit')
        ret.raise_for_status()
        #logging.info('ret: %r', ret.text)
        doc = BeautifulSoup(ret.text, 'html.parser')
        xsrf = doc.find('input', {'name': '_xsrf'}).get('value')  # type: ignore
        logging.info('xsrf: %r', xsrf)
        logging.info('cookies: %r', http_client.cookies)

        config_str = json.dumps(config)

        submit_mock = client.req_mock.add_mock('/actions/submit', MockError(400, '{"error": "Validation error"}'))

        ret = await client.request_raw(http_client, 'POST', '/submit', form_data={
            '_xsrf': xsrf,
            'submit_box': config_str,
            'description': description,
            'number_jobs': 10,
            'group': 'users',
        })
        assert ret.status_code == 400
        doc = BeautifulSoup(ret.text, 'html.parser')
        error = doc.find('div', {'id': 'error'}).text  # type: ignore
        assert 'Validation error' in error


async def test_website_submit_status_queued(server):
    client = server(username='username', roles=['user'], groups=['users', 'simprod'])

    description = 'Test dataset'

    config = Config({
        'tasks':[{
            'name': 'testing',
            'trays': [{
                'modules': [{
                    'src': '/usr/bin/python3',
                    'args': ''
                }]
            }]
        }],
        'version': 3.2,
    })
    config.fill_defaults()
    config.validate()
    config_str = json.dumps(config.config)

    async with client.get_http_client() as http_client:
        submit_mock = client.req_mock.add_mock('/actions/submit/123', {
            'status': 'queued',
            'payload': {
                'config': config_str,
                'description': description,
                'jobs_submitted': 10,
                'group': 'users',
            },
        })

        ret = await client.request_raw(http_client, 'GET', '/submit/status/123')
        ret.raise_for_status()

        assert ret.status_code == 200


async def test_website_submit_status_error(server):
    client = server(username='username', roles=['user'], groups=['users', 'simprod'])

    description = 'Test dataset'

    config = Config({
        'tasks':[{
            'name': 'testing',
            'trays': [{
                'modules': [{
                    'src': '/usr/bin/python3',
                    'args': ''
                }]
            }]
        }],
        'version': 3.2,
    })
    config.fill_defaults()
    config.validate()
    config_str = json.dumps(config.config)

    async with client.get_http_client() as http_client:
        submit_mock = client.req_mock.add_mock('/actions/submit/123', {
            'status': 'error',
            'payload': {
                'config': config_str,
                'description': description,
                'jobs_submitted': 10,
                'group': 'users',
            },
            'error_message': 'Validation error'
        })

        ret = await client.request_raw(http_client, 'GET', '/submit/status/123')
        ret.raise_for_status()

        assert ret.status_code == 200
        doc = BeautifulSoup(ret.text, 'html.parser')
        logging.info('doc: %s', ret.text)
        error = doc.find('div', {'id': 'error'}).text  # type: ignore
        assert error == 'Validation error'
        

async def test_website_submit_status_complete(server):
    client = server(username='username', roles=['user'], groups=['users', 'simprod'])

    description = 'Test dataset'

    config = Config({
        'tasks':[{
            'name': 'testing',
            'trays': [{
                'modules': [{
                    'src': '/usr/bin/python3',
                    'args': ''
                }]
            }]
        }],
        'version': 3.2,
    })
    config.fill_defaults()
    config.validate()
    config_str = json.dumps(config.config)

    async with client.get_http_client() as http_client:
        submit_mock = client.req_mock.add_mock('/actions/submit/123', {
            'status': 'complete',
            'payload': {
                'config': config_str,
                'description': description,
                'jobs_submitted': 10,
                'group': 'users',
                'dataset_id': '456',
            },
        })

        ret = await client.request_raw(http_client, 'GET', '/submit/status/123')
        logging.info('headers %r', ret.headers)

        assert ret.status_code == 303
        assert ret.headers['location'].endswith('/dataset/456')


async def test_website_config(server):
    client = server(username='username', roles=['user'], groups=['users', 'simprod'])

    client.req_mock.add_mock('/datasets/123', {})
    client.req_mock.add_mock('/config/123', {})

    ret = await client.request('GET', '/config', {'dataset_id': '123'})


async def test_website_config_status_queued(server):
    client = server(username='username', roles=['user'], groups=['users', 'simprod'])

    dataset_id = 'd1'
    description = 'Test dataset'

    config = Config({
        'tasks':[{
            'name': 'testing',
            'trays': [{
                'modules': [{
                    'src': '/usr/bin/python3',
                    'args': ''
                }]
            }]
        }],
        'version': 3.2,
    })
    config.fill_defaults()
    config.validate()
    config_str = json.dumps(config.config)

    async with client.get_http_client() as http_client:
        client.req_mock.add_mock('/actions/edit_config/123', {
            'status': 'queued',
            'payload': {
                'dataset_id': dataset_id,
                'config': config_str,
                'description': description,
            },
        })

        ret = await client.request_raw(http_client, 'GET', '/config/status/123')
        ret.raise_for_status()

        assert ret.status_code == 200


async def test_website_profile(server):
    client = server(username='username', roles=['user'], groups=['users', 'simprod'])

    ret = await client.request('GET', '/profile')








# """
# Test script for the website module.
# """

# from __future__ import absolute_import, division, print_function

# from tests.util import unittest_reporter, glob_tests, services_mock

# import logging
# logger = logging.getLogger('modules_website_test')

# import os, sys, time, random
# import shutil
# import tempfile
# import random
# import threading
# import subprocess
# from datetime import datetime,timedelta
# from functools import partial
# import unittest
# from contextlib import contextmanager
# import urllib

# from unittest.mock import patch, MagicMock

# import asyncio

# import tornado.ioloop
# from tornado.testing import AsyncTestCase
# from tornado.httpclient import AsyncHTTPClient, HTTPError


# from bs4 import BeautifulSoup

# import iceprod.core.logger
# from iceprod.core import to_log
# from iceprod.core.jsonUtil import json_encode,json_decode
# import iceprod.server
# from iceprod.server.modules.website import website
# import iceprod.server.modules.website
# import iceprod.server.module
# try:
    # from iceprod.server import ssl_cert
# except ImportError:
    # ssl_cert = None

# from ..module_test import module_test

# # check for javascript testing
# try:
    # from selenium import webdriver
    # from selenium.webdriver.support.ui import WebDriverWait
    # from selenium.webdriver.support import expected_conditions as EC
    # with to_log(sys.stdout):
        # subprocess.check_call(['which','phantomjs'])
    # testjs = True
# except (ImportError, subprocess.CalledProcessError):
    # logger.info('skipping javascript tests', exc_info=True)
    # testjs = False


# class website_test(AsyncTestCase):
    # def setUp(self):
        # super(website_test,self).setUp()

        # orig_dir = os.getcwd()
        # self.test_dir = tempfile.mkdtemp(dir=orig_dir)
        # os.chdir(self.test_dir)
        # def clean_dir():
            # os.chdir(orig_dir)
            # shutil.rmtree(self.test_dir)
        # self.addCleanup(clean_dir)

        # try:
            # # set hostname
            # self.hostname = b'localhost'

            # self.cfg = {'webserver':{'port':random.randint(10000,32000)},
                        # 'db':{'name':'test'},
                        # 'site_id':'abc',
                        # 'download':{'http_username':None,
                                    # 'http_password':None,
                                   # },
                        # 'rest_api':{'url':'foo','auth_key':'bar'},
                       # }
            # self.executor = {}
            # self.modules = services_mock()
            
            # self.website = website(self.cfg, self.executor, self.modules)
        # except:
            # logger.warning('error setting up modules_website', exc_info=True)
            # raise

        # self.auth_cookie = None

    # @contextmanager
    # def start(self):
        # try:
            # self.website.start()
            # yield
        # finally:
            # self.website.stop()

    # def request(self, *args, **kwargs):
        # if self.auth_cookie:
            # if 'headers' in kwargs:
                # kwargs['headers']['cookie'] = self.auth_cookie
            # else:
                # kwargs['headers'] = {'cookie': self.auth_cookie}
        # if 'follow_redirects' not in kwargs:
            # kwargs['follow_redirects'] = False
        # client = AsyncHTTPClient()
        # return client.fetch(*args, **kwargs)


    # @unittest_reporter(name='start/stop/kill')
    # async def test_010_start_stop_kill(self):
        # self.website.start()

        # await self.website.stop()

        # self.website.start()
        # self.website.kill()

    # @patch('iceprod.server.modules.website.RestClient')
    # @unittest_reporter(name=' /')
    # async def test_110_Default(self, req):
        # with self.start():
            # datasets_status = {'processing':['d1','d2']}

            # address = 'http://localhost:%d'%(self.cfg['webserver']['port'])
            # logger.info('try connecting at %s',address)

            # await self.get_auth_cookie()

            # async def rest(method, url, args=None):
                # if url.startswith('/dataset_summaries/status'):
                    # rest.called = True
                    # return datasets_status
                # else:
                    # raise Exception()
            # rest.called = False
            # req.return_value.request = rest

            # # main site
            # r = await self.request(address)
            # self.assertEqual(r.code, 200)
            # body = r.body.decode('utf-8')
            # self.assertFalse(rest.called)

    # @patch('iceprod.server.modules.website.RestClient')
    # @unittest_reporter(name=' /submit')
    # async def test_200_Submit(self, req):
        # with self.start():
            # address = 'http://localhost:%d/submit'%(self.cfg['webserver']['port'])
            # logger.info('try connecting at %s',address)

            # await self.get_auth_cookie()

            # async def rest(method, url, args=None):
                # if url.startswith('/create_token'):
                    # rest.called = True
                    # return {'result': 'token'}
                # else:
                    # raise Exception()
            # rest.called = False
            # req.return_value.request = rest

            # r = await self.request(address)
            # self.assertEqual(r.code, 200)
            # self.assertTrue(rest.called)

    # @patch('iceprod.server.modules.website.RestClient')
    # @unittest_reporter(name=' /config')
    # async def test_210_Config(self, req):
        # with self.start():
            # dataset_id = 'foo'
            # address = 'http://localhost:{}/config?dataset_id={}'.format(
                # self.cfg['webserver']['port'], dataset_id)
            # logger.info('try connecting at %s',address)

            # await self.get_auth_cookie()

            # async def rest(method, url, args=None):
                # if url.startswith('/datasets/{}'.format(dataset_id)):
                    # return {'dataset_id':'foo'}
                # elif url.startswith('/config/{}'.format(dataset_id)):
                    # rest.called = True
                    # return {'dataset_id':'foo'}
                # else:
                    # raise Exception()
            # rest.called = False
            # req.return_value.request = rest

            # r = await self.request(address)
            # self.assertEqual(r.code, 200)
            # self.assertTrue(rest.called)
            
            # address = 'http://localhost:{}/config?dataset_id={}&edit=1'.format(
                # self.cfg['webserver']['port'], dataset_id)
            # logger.info('try connecting at %s',address)

            # async def rest(method, url, args=None):
                # if url.startswith('/create_token'):
                    # rest.called = True
                    # return {'result': 'token'}
                # elif url.startswith('/datasets/{}'.format(dataset_id)):
                    # return {'dataset_id':'foo'}
                # elif url.startswith('/config/{}'.format(dataset_id)):
                    # return {'dataset_id':'foo'}
                # else:
                    # raise Exception()
            # rest.called = False
            # req.return_value.request = rest

            # r = await self.request(address)
            # self.assertEqual(r.code, 200)
            # self.assertTrue(rest.called)
            
            # address = 'http://localhost:{}/config'.format(
                # self.cfg['webserver']['port'])
            # logger.info('try connecting at %s',address)

            # r = await self.request(address, raise_error=False)
            # self.assertEqual(r.code, 400)

    # @patch('iceprod.server.modules.website.RestClient')
    # @unittest_reporter(name=' /dataset')
    # async def test_300_Dataset(self, req):
        # with self.start():
            # dataset_id = 'foo'
            # address = 'http://localhost:{}/dataset'.format(
                # self.cfg['webserver']['port'])
            # logger.info('try connecting at %s',address)

            # await self.get_auth_cookie()

            # async def rest(method, url, args=None):
                # if url.startswith('/datasets'):
                    # rest.called = True
                    # return {dataset_id:{'dataset_id':dataset_id, 'status':'processing',
                                        # 'dataset': 1234, 'description':'desc'}}
                # else:
                    # raise Exception()
            # rest.called = False
            # req.return_value.request = rest

            # r = await self.request(address)
            # self.assertEqual(r.code, 200)
            # self.assertTrue(rest.called)

    # @patch('iceprod.server.modules.website.RestClient')
    # @unittest_reporter(name=' /dataset/<dataset_id>')
    # async def test_301_Dataset(self, req):
        # with self.start():
            # dataset_id = 'foo'
            # address = 'http://localhost:{}/dataset/{}'.format(
                # self.cfg['webserver']['port'], dataset_id)
            # logger.info('try connecting at %s',address)

            # await self.get_auth_cookie()

            # async def rest(method, url, args=None):
                # if url == '/datasets/{}'.format(dataset_id):
                    # rest.called = True
                    # return {'dataset_id':dataset_id, 'status':'processing',
                            # 'dataset': 1234, 'description':'desc'}
                # elif url.startswith('/datasets/{}/job_counts/status'.format(dataset_id)):
                    # return {'processing':['foo','bar']}
                # elif url.startswith('/datasets/{}/task_counts/status'.format(dataset_id)):
                    # return {'processing':['1','2','3'], 'complete': ['4','5'], 'suspended':['6']}
                # elif url.startswith('/datasets/{}/task_counts/name_status'.format(dataset_id)):
                    # return {'generator':{'complete':2},
                            # 'hits':{'processing':1,'reset':1},
                            # 'detector':{'waiting':1,'suspended':1},
                           # }
                # elif url.startswith('/datasets/{}/task_stats'.format(dataset_id)):
                    # return {'generator':{'count': 2, 'gpu': 0, 'total_hrs': 1.2, 'total_err_hrs': 3.4,
                                         # 'avg_hrs': 0.7, 'stddev_hrs': 0.3, 'min_hrs': 0.4,
                                         # 'max_hrs': 1.1, 'efficiency': 0.54},
                            # 'hits':{'count': 2, 'gpu': 2, 'total_hrs': 1.2, 'total_err_hrs': 3.4,
                                         # 'avg_hrs': 0.7, 'stddev_hrs': 0.3, 'min_hrs': 0.4,
                                         # 'max_hrs': 1.1, 'efficiency': 0.54},
                            # 'detector':{'count': 2, 'gpu': 0, 'total_hrs': 1.2, 'total_err_hrs': 3.4,
                                         # 'avg_hrs': 0.7, 'stddev_hrs': 0.3, 'min_hrs': 0.4,
                                         # 'max_hrs': 1.1, 'efficiency': 0.54},
                           # }
                # elif url.startswith('/create_token'):
                    # return {'result': 'token'}
                # elif url == f'/config/{dataset_id}':
                    # return {'tasks':[]}
                # else:
                    # logger.info(f'{method} {url}')
                    # raise Exception()
            # rest.called = False
            # req.return_value.request = rest

            # r = await self.request(address)
            # self.assertEqual(r.code, 200)
            # self.assertTrue(rest.called)

            # async def rest(method, url, args=None):
                # raise Exception()
            # req.return_value.request = rest
            # r = await self.request(address, raise_error=False)
            # self.assertEqual(r.code, 404)

    # @patch('iceprod.server.modules.website.RestClient')
    # @patch('iceprod.server.GlobalID.globalID_gen')
    # @unittest_reporter(name=' /dataset/<dataset_num>')
    # async def test_302_Dataset(self, globalid, req):
        # with self.start():
            # dataset_id = 'foo'
            # dataset_num = 120
            # address = 'http://localhost:{}/dataset/{}'.format(
                # self.cfg['webserver']['port'], dataset_num)
            # logger.info('try connecting at %s',address)

            # await self.get_auth_cookie()

            # globalid.return_value = dataset_id

            # async def rest(method, url, args=None):
                # logger.info('REST %s %s', method, url)
                # if url == '/datasets/{}'.format(dataset_id):
                    # rest.called = True
                    # return {'dataset_id':dataset_id, 'status':'processing',
                            # 'dataset': dataset_num, 'description':'desc'}
                # elif url == '/datasets':
                    # return {dataset_id: {'dataset_id':dataset_id, 'status':'processing',
                            # 'dataset': dataset_num, 'description':'desc'}}
                # elif url.startswith('/datasets/{}/job_counts/status'.format(dataset_id)):
                    # return {'processing':['foo','bar']}
                # elif url.startswith('/datasets/{}/task_counts/status'.format(dataset_id)):
                    # return {'processing':['1','2','3'], 'complete': ['4','5'], 'suspended':['6']}
                # elif url.startswith('/datasets/{}/task_counts/name_status'.format(dataset_id)):
                    # return {'generator':{'complete':2},
                            # 'hits':{'processing':1,'reset':1},
                            # 'detector':{'waiting':1,'suspended':1},
                           # }
                # elif url.startswith('/datasets/{}/task_stats'.format(dataset_id)):
                    # return {'generator':{'count': 2, 'gpu': 0, 'total_hrs': 1.2, 'total_err_hrs': 3.4,
                                         # 'avg_hrs': 0.7, 'stddev_hrs': 0.3, 'min_hrs': 0.4,
                                         # 'max_hrs': 1.1, 'efficiency': 0.54},
                            # 'hits':{'count': 2, 'gpu': 2, 'total_hrs': 1.2, 'total_err_hrs': 3.4,
                                         # 'avg_hrs': 0.7, 'stddev_hrs': 0.3, 'min_hrs': 0.4,
                                         # 'max_hrs': 1.1, 'efficiency': 0.54},
                            # 'detector':{'count': 2, 'gpu': 0, 'total_hrs': 1.2, 'total_err_hrs': 3.4,
                                         # 'avg_hrs': 0.7, 'stddev_hrs': 0.3, 'min_hrs': 0.4,
                                         # 'max_hrs': 1.1, 'efficiency': 0.54},
                           # }
                # elif url.startswith('/create_token'):
                    # return {'result': 'token'}
                # elif url == f'/config/{dataset_id}':
                    # return {'tasks':[]}
                # else:
                    # raise Exception()
            # rest.called = False
            # req.return_value.request = rest

            # r = await self.request(address)
            # self.assertEqual(r.code, 200)
            # self.assertTrue(rest.called)

            # async def rest(method, url, args=None):
                # raise Exception()
            # req.return_value.request = rest
            # r = await self.request(address, raise_error=False)
            # self.assertEqual(r.code, 404)

    # @patch('iceprod.server.modules.website.RestClient')
    # @unittest_reporter(name=' /dataset/<dataset_id>/task')
    # async def test_400_Task(self, req):
        # with self.start():
            # dataset_id = 'foo'
            # task_id = 'bar'
            # address = 'http://localhost:{}/dataset/{}/task'.format(
                # self.cfg['webserver']['port'], dataset_id)
            # logger.info('try connecting at %s',address)

            # await self.get_auth_cookie()

            # async def rest(method, url, args=None):
                # if url.startswith('/datasets/{}/task_counts/status'.format(dataset_id)):
                    # rest.called = True
                    # return {'processing':['1','2','3'], 'complete': ['4','5'], 'suspended':['6']}
                # else:
                    # raise Exception()
            # rest.called = False
            # req.return_value.request = rest

            # r = await self.request(address)
            # self.assertEqual(r.code, 200)
            # self.assertTrue(rest.called)
            
            # address = 'http://localhost:{}/dataset/{}/task?status=processing'.format(
                # self.cfg['webserver']['port'], dataset_id)

            # async def rest(method, url, args=None):
                # logger.info('REST %s %s', method, url)
                # if url.startswith('/datasets/{}/tasks'.format(dataset_id)):
                    # rest.called = True
                    # return {task_id:{'task_id':task_id,'dataset_id':dataset_id,
                                     # 'task_index': 0, 'job_index': 0,
                                     # 'job_id':'j', 'status':'processing',
                                     # 'name':'generator', 'failures':0}}
                # elif url.startswith('/datasets/{}/job'.format(dataset_id)):
                    # return {'job_index':0}
                # elif url.startswith('/create_token'):
                    # return {'result': 'token'}
                # else:
                    # raise Exception()
            # rest.called = False
            # req.return_value.request = rest

            # r = await self.request(address)
            # self.assertEqual(r.code, 200)
            # self.assertTrue(rest.called)
            # soup = BeautifulSoup(r.body, "html.parser")
            # for e in soup.findAll("a"):
                # if e.get('href').endswith('/bar'):
                    # break
            # else:
                # raise Exception('did not find task_id bar')

            # async def rest(method, url, args=None):
                # logger.info('REST %s %s', method, url)
                # if url.startswith('/datasets/{}/tasks'.format(dataset_id)):
                    # rest.called = True
                    # return {}
                # elif url.startswith('/create_token'):
                    # return {'result': 'token'}
                # else:
                    # raise Exception()
            # rest.called = False
            # req.return_value.request = rest
                
            # address = 'http://localhost:{}/dataset/{}/task?status=waiting'.format(
                # self.cfg['webserver']['port'], dataset_id)
            # r = await self.request(address)
            # self.assertEqual(r.code, 200)
            # self.assertTrue(rest.called)
            # soup = BeautifulSoup(r.body, "html.parser")
            # for e in soup.findAll("a"):
                # self.assertFalse(e.get('href').endswith('/bar'))

    # @patch('iceprod.server.modules.website.RestClient')
    # @unittest_reporter(name=' /dataset/<dataset_id>/task/<task_id>')
    # async def test_401_Task(self, req):
        # with self.start():
            # dataset_id = 'foo'
            # task_id = 'bar'
            # address = 'http://localhost:{}/dataset/{}/task/{}'.format(
                # self.cfg['webserver']['port'], dataset_id, task_id)
            # logger.info('try connecting at %s',address)

            # await self.get_auth_cookie()

            # async def rest(method, url, args=None):
                # logger.info('REST %s %s', method, url)
                # if url == '/datasets/{}'.format(dataset_id):
                    # return {'dataset_id':dataset_id, 'status':'processing',
                            # 'dataset': 1000, 'description':'desc'}
                # elif url == '/datasets/{}/tasks/{}'.format(dataset_id,task_id):
                    # rest.called = True
                    # return {'task_id':task_id,'dataset_id':dataset_id,
                            # 'task_index': 0, 'job_index': 0,
                            # 'job_id':'j', 'status':'processing',
                            # 'name':'generator', 'failures':0}
                # elif url.startswith('/datasets/{}/tasks/{}/log'.format(dataset_id,task_id)):
                    # return {'logs':[{'log_id':'baz','name':'stdout','data':'foo\nbar',
                                     # 'dataset_id':dataset_id,'task_id':task_id}]}
                # elif url.startswith('/datasets/{}/tasks/{}/task_stats'.format(dataset_id,task_id)):
                    # return {}
                # elif url.startswith('/create_token'):
                    # return {'result': 'token'}
                # else:
                    # raise Exception()
            # rest.called = False
            # req.return_value.request = rest

            # r = await self.request(address)
            # self.assertEqual(r.code, 200)
            # self.assertTrue(rest.called)

    # @patch('iceprod.server.modules.website.RestClient')
    # @unittest_reporter(name=' /dataset/<dataset_id>/job')
    # async def test_500_Job(self, req):
        # with self.start():
            # dataset_id = 'foo'
            # job_id = 'bar'
            # address = 'http://localhost:{}/dataset/{}/job'.format(
                # self.cfg['webserver']['port'], dataset_id)
            # logger.info('try connecting at %s',address)

            # await self.get_auth_cookie()

            # async def rest(method, url, args=None):
                # logger.info('REST %s %s', method, url)
                # if url.startswith('/datasets/{}/jobs'.format(dataset_id)):
                    # rest.called = True
                    # return {job_id:{'job_id':job_id,'dataset_id':dataset_id,
                                    # 'status':'processing', 'status_changed':'2018',
                                    # 'job_index':0}}
                # elif url.startswith('/create_token'):
                    # return {'result': 'token'}
                # else:
                    # raise Exception()
            # rest.called = False
            # req.return_value.request = rest

            # r = await self.request(address)
            # self.assertEqual(r.code, 200)
            # self.assertTrue(rest.called)

            # address = 'http://localhost:{}/dataset/{}/job?status=processing'.format(
                # self.cfg['webserver']['port'], dataset_id)

            # r = await self.request(address)
            # self.assertEqual(r.code, 200)
            # soup = BeautifulSoup(r.body, "html.parser")
            # for e in soup.findAll("a"):
                # if e.get('href').endswith('/bar'):
                    # break
            # else:
                # raise Exception('did not find task_id bar')
                
            # address = 'http://localhost:{}/dataset/{}/job?status=waiting'.format(
                # self.cfg['webserver']['port'], dataset_id)
            # r = await self.request(address)
            # self.assertEqual(r.code, 200)
            # soup = BeautifulSoup(r.body, "html.parser")
            # for e in soup.findAll("a"):
                # self.assertFalse(e.get('href').endswith('/bar'))

    # @patch('iceprod.server.modules.website.RestClient')
    # @unittest_reporter(name=' /dataset/<dataset_id>/job/<job_id>')
    # async def test_501_Job(self, req):
        # with self.start():
            # dataset_id = 'foo'
            # job_id = 'bar'
            # address = 'http://localhost:{}/dataset/{}/job/{}'.format(
                # self.cfg['webserver']['port'], dataset_id, job_id)
            # logger.info('try connecting at %s',address)

            # await self.get_auth_cookie()

            # async def rest(method, url, args=None):
                # logger.info('REST %s %s', method, url)
                # if url == '/datasets/{}'.format(dataset_id):
                    # return {'dataset_id':dataset_id, 'status':'processing',
                            # 'dataset': 1000, 'description':'desc'}
                # elif url.startswith('/datasets/{}/jobs/{}'.format(dataset_id,job_id)):
                    # rest.called = True
                    # return {'job_id':job_id,'dataset_id':dataset_id,
                            # 'status':'processing', 'status_changed':'2018',
                            # 'job_index':0}
                # elif url.startswith('/datasets/{}/tasks'.format(dataset_id)):
                    # return {'baz':{'task_id':'baz','dataset_id':dataset_id,
                                   # 'job_id':job_id,'name':'generate',
                                   # 'status':'queued', 'status_changed':'2018',
                                   # 'task_index':0, 'failures':0, 'requirements':{'cpu':1},
                                   # 'walltime':0.,'walltime_err':0.}}
                # elif url.startswith('/create_token'):
                    # return {'result': 'token'}
                # else:
                    # raise Exception()
            # rest.called = False
            # req.return_value.request = rest

            # r = await self.request(address)
            # self.assertEqual(r.code, 200)
            # self.assertTrue(rest.called)

    # @patch('iceprod.server.modules.website.RestClient')
    # @unittest_reporter(name=' /dataset/<dataset_id/log/<log_id>')
    # async def test_600_Log(self, req):
        # with self.start():
            # dataset_id = 'foo'
            # log_id = 'bar'
            # address = 'http://localhost:{}/dataset/{}/log/{}'.format(
                # self.cfg['webserver']['port'], dataset_id, log_id)
            # logger.info('try connecting at %s',address)

            # await self.get_auth_cookie()

            # async def rest(method, url, args=None):
                # logger.info('REST %s %s', method, url)
                # if url.startswith('/datasets/{}/logs/{}'.format(dataset_id,log_id)):
                    # rest.called = True
                    # return {'log_id':log_id,'dataset_id':dataset_id,
                            # 'name':'name', 'data': 'this is a log\nfoo bar'}
                # else:
                    # raise Exception()
            # rest.called = False
            # req.return_value.request = rest

            # r = await self.request(address)
            # self.assertEqual(r.code, 200)
            # self.assertTrue(rest.called)
            # self.assertIn(b'this is a log', r.body)

